# -*- coding: utf-8 -*-

# module Unit

# Copyright (C) 2019 National Research Council Canada
# Author:  Harold Parks

# This file is part of MetroloPy.

# MetroloPy is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.

# MetroloPy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more 
# details.

# You should have received a copy of the GNU General Public License along with 
# MetroloPy. If not, see <http://www.gnu.org/licenses/>.

"""
This module contains the procedures for defining units and unit algebera used
by the gummy class.
"""

import numpy as np
import weakref
import warnings
from importlib import import_module
from .exceptions import (UnitLibError,CircularUnitConversionError,
                         NoUnitConversionFoundError,UnitNotFoundError,
                         IncompatibleUnitsError,UnitLibNotFoundError,
                         UnitWarning)
from .printing import PrettyPrinter
from .nummy import nummy

class _UnitParser:
    def __init__(self,txt):  
        self.txt= txt
        self.n = len(txt)
        self.i = 0
        
    def parse(self):
        return self.ppar(0,1)
                
    def ppar(self,lev,es):
        cu = ''
        ul = []
        sol = False
        while self.i < self.n:
            c = self.txt[self.i]
            if c == ')':
                cu = cu.strip()
                if cu != '':
                    ul += [[cu,es]]
                if len(ul) == 0:
                    raise UnitLibError('syntax error in: ' + self.txt)
                self.i += 1
                e = self.look_for_pow()
                if e != 1:
                    for u in ul:
                        u[1] *= e
                        if np.modf(u[1]) == 0:
                            u[1] = int(u[1])
                return ul
            elif c == '(':
                cu = cu.strip()
                p = self.pch(cu,')',']',self.i)
                if p is None:
                    if cu != '':
                        ul += [[cu,es]]
                        cu = ''
                    self.i += 1
                    ul += self.ppar(lev+1,es)
                    if sol:
                        es = -es
                        sol = False
                else:
                    cu += p
                    self.i += len(p)
            elif c == '[':
                cu = cu.strip()
                p = self.pch(cu,']',')',self.i)
                if p is None:
                    if cu != '':
                        ul += [[cu,es]]
                        cu = ''
                    self.i += 1
                    ul += [self.pbr(es,cu)]
                    cu = ''
                    if sol:
                        es = -es
                        sol = False
                else:
                    cu += p
                    self.i += len(p)
            elif c.isspace() or c == '*':
                cu = cu.strip()
                if cu != '':
                    ul += [[cu,es*self.look_for_pow()]]
                cu = ''
                if sol:
                    es = -es
                    sol = False
            elif c == '/':
                cu = cu.strip()
                if cu != '':
                    ul += [[cu,es]]
                    cu = ''
                 #if sol:
                    # raise UnitLibError('ambugus use of several soliudses (forward slashes) in: ' + self.txt)
                sol = not sol
                es = -es
                self.i += 1
            else:
                cu += c
                self.i += 1
        if lev != 0:
            raise UnitLibError('unmatched ")" found in: ' + self.txt)
        cu = cu.strip()
        if cu != '':
            ul += [[cu,es]]
        return ul
            
    def look_for_pow(self):
        while self.i < self.n:
            c = self.txt[self.i]
            if c == '*':
                if self.i == (self.n - 1):
                    raise UnitLibError(self.txt + ' ends with *')
                self.i += 1
                c = self.txt[self.i]
                if not c == '*':
                    return 1
                par = -1
                et = ''
                self.i += 1
                cl = ''
                stnum = False
                while self.i < self.n:
                    c = self.txt[self.i]
                    if c == '(':
                        if par == 0:
                            if cl.ispace() or cl == '*':
                                f =  float(et)
                                if np.modf(f)[0] == 0:
                                    return int(f)
                                else:
                                    return f
                            raise UnitLibError('syntax error in: ' + self.txt)
                        else:
                            if par == -1:
                                par = 1
                            else:
                                par += 1
                    elif c == '[':
                        if cl.ispace() or cl == '*':
                            f =  float(et)
                            if np.modf(f)[0] == 0:
                                return int(f)
                            else:
                                return f
                        raise UnitLibError('syntax error in: ' + self.txt )
                    elif c == ']':
                        raise UnitLibError('unmatched bracket: ' + self.txt)
                    elif c.isalpha():
                        if par > 0:
                            raise UnitLibError('syntax error in: ' + self.txt)
                        f =  float(et)
                        if np.modf(f)[0] == 0:
                            return int(f)
                        else:
                            return f
                    elif c == '/':
                        if par <= 0:
                            f =  float(et)
                            if np.modf(f)[0] == 0:
                                return int(f)
                            else:
                                return f
                    elif c == ')':
                        if par <= 0:
                            f =  float(et)
                            if np.modf(f)[0] == 0:
                                return int(f)
                            else:
                                return f
                        else:
                            par -= 1
                    elif c.isnumeric():
                        stnum = True
                    elif c == '+'  or c == '-':
                        if stnum and par <= 0:
                            raise UnitLibError('syntax error in: ' + self.txt)
                    elif c == '*':
                        if self.i == self.n - 1:
                            raise UnitLibError('syntax error in: ' + self.txt)
                        if par <= 0 and cl != '*' and self.txt[self.i+1] != '*':
                            self.i += 1
                            f =  float(et)
                            if np.modf(f)[0] == 0:
                                return int(f)
                            else:
                                return f
                    elif c.isspace():
                        if stnum and par <= 0:
                            self.i += 1
                            f =  float(et)
                            if np.modf(f)[0] == 0:
                                return int(f)
                            else:
                                return f
                    else:
                        if par <= 0:
                            raise UnitLibError('syntax error in: ' + self.txt)
                    self.i += 1
                    cl = c
                    et += c
                f = float(et)
                if np.modf(f)[0] == 0:
                    return int(f)
                else:
                    return f
    
            elif not c.isspace():
                return 1
            self.i += 1
        return 1
            
    def pbr(self,es,cu):
        br = 0
        while self.i < self.n:
            c = self.txt[self.i]
            if c == '[':
                br += 1
                cu += c
            if c == ']':
                if br == 0:
                    self.i += 1
                    return [cu,es*self.look_for_pow()]
                br -= 1
            cu += c
            self.i += 1
        raise UnitLibError('syntax error in: ' + self.txt)
        
    def pch(self,cu,ech,och,i):
        if cu == '':
            return None
        ret = self.txt[i]
        i += 1
        while i < self.n:
            c = self.txt[i]
            
            if c == ech:
                return ret + c
            if c.isspace():
                return None
            if c in {och,'*','/'}:
                return None
            
            if c == '(':
                r = self.pch('a',')',']',i)
                if r is None:
                    return None
                else:
                    ret += r
                    i += len(r) - 1
            elif c == '[':
                r = self.pch('a',']',')',i)
                if r is None:
                    return None
                else:
                    ret += r
                    i += len(r) - 1
            else:
                ret += c
                
            i += 1
            
        return None
        
        
class Conversion:
    """
    Represents a unit conversion.  This class should only be used as arguments
    to `Unit` object initializers.  Each conversion should be associated with one
    and only one parent Unit.
    
    Parameters
    ----------
    unit:  `str` or `Unit`
        the Unit that the parent Unit will be converted to.
        
    factor:  `float`, optional
        The conversion factor between the parent Unit and the new Unit:
        [value with new Unit] = factor * [value with parent Unit]
        The default value is 1
    """
    linear = True
    
    def __init__(self,unit,factor=1):
        self._unit = unit
        self.factor = factor
        
    def _to(self,g):
        return g*self.factor
        
    def to(self,g):
        unit_in = None
        try:
            try:
                unit_in = g._unit
                g._unit = one
            except:
                pass
            ret = self._to(g)
            try:
                if unit_in is not None:
                    ret._unit = self.unit
            except:
                pass
            return ret
        finally:
            if unit_in is not None:
                try:
                    g._unit = unit_in
                except:
                    pass
    
    def _frm(self,g):
        return g/self.factor
        
    def frm(self,g):
        unit_in = None
        try:
            try:
                unit_in = g._unit
                g._unit = one
            except:
                pass
            ret = self._frm(g)
            try:
                if unit_in is not None:
                    ret._unit = self.parent
            except:
                pass
            return ret
        finally:
            if unit_in is not None:
                try:
                    g._unit = unit_in
                except:
                    pass
        
    def chain(self,c):
        if c.linear:
            ret = Conversion(c.unit,self.factor*c.factor)
            ret.parent = self.parent
            return ret
        else:
            return c.rchain(self)
        
    @property
    def unit(self):
        if not isinstance(self._unit,Unit):
            self._unit = Unit.unit(self._unit)
        return self._unit
        
    def copy(self):
        r = Conversion(self.unit,self.factor)
        r.parent = self.parent
        return r
        
    def pow(self,e):
        r = Conversion(self.unit**e,self.factor**e)
        r.parent = self.parent**e
        return r


class Unit(PrettyPrinter):
    """
    Creating an instance of this class creates a representation of a physical
    unit and adds it to the unit library.  Units already in the unit library or 
    derived units made up of other units in the unit library can be accessed 
    by passing a text string with the unit name or symbol to the static method
    `Unit.unit`.
    
    Parameters
    ----------
        
    name:  `str`
        The name of the unit.  The name can be used access the unit with the
        `Unit.unit` method, but note that if you define a Unit with an
        identical name to a previously defined unit then the older name will
        be shadowed.
    
    symbol:  `str`
        A unicode symbol used when displaying the unit.  If the `add_symbol`
        parameter is set to `True`, then this symbol can also be used
        to access the unit with the `Unit.unit` method.
    
    conversion: `Conversion` or `None`, optional
        A conversion to another unit.  When creating units be careful to avoid
        circular conversions, i.e. you can define:
            
        Unit('inch','in',conversion=None)
        Unit('foot','ft',conversion=Conversion('in',12))
        Unit('yard','yd',conversion=Conversion('ft',3))
        
        but not:
            
        Unit('inch','in',conversion=Conversion('yd',1/36))
        Unit('foot','ft',conversion=Conversion('in',12))
        Unit('yard','yd',conversion=Conversion('ft',3))
        
        Note that an equivent and allowable way of defining the first set of 
        units above is:
            
        Unit('inch','in',conversion=None)
        Unit('foot','ft',conversion=Conversion('in',12))
        Unit('yard','yd',conversion=Conversion('in',36))
        
        Either way will allow the free conversion between inches, feet, and
        yards.  You can also define the inch as:
            
        Unit('inch','in',conversion=('cm',2.54)
        
        (The inch is actually defined this way in the builtin unit library.)
    
    short_name: `str` or `None`
        a short name
    
    add_symbol: `bool`, optional
        If this is `True`, then the symbol can be used to look up the unit
        in the unit library.  The default is `False`
        
    html_symbol, latex_symbol, ascii_symbol:  `str` or `Mone`, optional
        html, latex, and ascii versions of the symbol if they are different
        from the unicode representation of the symbol.
        
    description:  `str` or `None`, optional
        a description of the unit
    
    order:  `int` , optional
        When displaying composite derived units, the symbols with a lower
        `order` value will be displayed.  The default if -1.
        
        
    See Also
    --------
    The following are sub-classes of `Unit`:

    PrefixedUnit:  Creates a set of units with SI prefixes (..., kilo, mega, giga, ...)
    
    BinaryPrefixedUnit:  Creates a set of unit with binary prefixes (..., kibi, mebi, gibi, ...)
    
    LogUnit:  Logrithmic units (e.g. decibel or neper)
    
    OffsetUnit:  Units with an offset origin (degree Celsius or degree Fahrenheit)
    """
    # to speed loading of the module, only wait to load units until they are
    # needed.  If a unit cannot be found the modules below are loaded in 
    # reverse order.
    _builtins_to_import = ['..usunits','..siunits','..relunits']
    
    # _lib and _builtin_lib are dictionarys with string keys that reference
    # Unit object instances.  _lib is searched first so definitions in _lib
    # may shadow definitions in _builtin_lib.
    
    _lib = {} # Contains Unit instances created or loaded by the user.
    _builtin_lib = {} # Contains Unit instances created in the gummy module.
    
    _open_lib = _lib
    _used_units = {}
        
    @staticmethod
    def unit(txt,exception=True):
        """
        Finds an returns a Unit object from the Unit library.
        
        Parameters:
            txt: `str`, `Unit` or 1
                This may be a string representing the unit.  The string can
                contain the name, short name or (if the unit was created
                with `add_symbol` set to `True`) the symbol of the unit or a
                combination of names and/or symbols of several different
                units.  Spaces or the character '*' represent multiplication,
                the character '/' represents division and the string '**'
                represents the power operator. For example txt can be:
                    
                'kg m**2/s'
                
                or equivalently:
                    
                'kilogram*metre*metre*second**-1' or '(kg/s)*m**2'.
                
                If a unit name contains a space, '*' or '/' character then the
                name must be enclosed in square brackets, e.g:
                    
                [light year]
                
                If txt is a Unit instance that instance is returned.
                
            exception: `bool`, optional
                If this is `True` then a `UnitNotFoundError` or UnitLibError
                is raised if a unit is not found that matches `txt`.  If it is
                `False` and a unit is not found, then `Unit.unit` returns
                `None` without raising an exception.  The default is `True`.
        """
        try:
            if isinstance(txt,Unit):
                return txt
            if txt is None or txt is 1:
                return one
            try:
                txt = txt.strip()
            except AttributeError:
                raise UnitNotFoundError('a unit may only be represented by a Unit object, the integer "1", or a string')
            if txt == '' or txt == '1':
                return one
            uu = Unit._used_units.get(txt)
            if uu is not None:
                return uu 
            ul = _UnitParser(txt).parse()
            if len(ul) == 1 and ul[0][1] == 1:
                ret = Unit._single_unit(ul[0][0])
                return ret
            ull = []
            for u in ul:
                un = Unit._single_unit(u[0])
                if hasattr(un,'_getme'):
                    un = un._getme(ul,u[1])
                if isinstance(un,_CompositeUnit):
                    ull += [(uu[0],uu[1]*u[1]) for uu in un._units]
                else:
                    ull.append((un,u[1]))
            ret = _CompositeUnit(ull)
            Unit._used_units[txt] = ret
            return ret
        except UnitNotFoundError or UnitLibError:
            if (not txt.startswith('[') and not txt.endswith(']')):
                return Unit.unit('[' + txt + ']')
            elif exception:
                raise
            return None
        
    @staticmethod
    def reorder(txt):
        """
        This changes the order in which the symbols of composite derived units
        are displayed.

        Examples
        --------
        
        >>> print(Unit.unit('ft lb'))
        ft lb
        >>> print(Unit.unit('lb ft'))  #This is the same unit as above and displays identically
        ft lb
        >>> Unit.reorder('lb ft')  #Now the order will be changed when the unit is displayed
        >>> print(Unit.unit('ft lb'))
        lb ft
            
        """
        unit = Unit.unit(txt)
        if not isinstance(unit,_CompositeUnit):
            return
        
        ul = _UnitParser(txt).parse()
        for u in ul:
            un = Unit._single_unit(u[0])
            if hasattr(un,'_getme'):
                u[0] = un._getme(ul,u[1])
            else:
                u[0] = un
            if isinstance(un,_CompositeUnit):
                raise ValueError('aliases for composite units may not be used in Unit.reorder')
        unit._make_symbols(ul)
                
    @staticmethod
    def _single_unit(txt):
        if txt == '1' or txt == '' or txt is 1:
            return one

        un = Unit._lib.get(txt)
        if un is None:
            un = Unit._builtin_lib.get(txt)
            if un is None:
                while len(Unit._builtins_to_import) > 0:
                    import_module(Unit._builtins_to_import.pop(),Unit.__module__)
                    un = Unit._builtin_lib.get(txt)
                    if un is not None:
                        break    

        if un is None:
            raise UnitNotFoundError('unit "' + txt + '" was not found')
        return un
        
    @staticmethod
    def alias(alias,unit):
        """
        Creates an alias that can be used to reference a Unit.
        
        Parameters
        ----------
        alias:  `str`
            a string containing the new alias

        unit:  `str` or `Unit`
            A string referencing the `Unit` that will be assigned
            the alias or the `Unit` instance its self.
        """
        unit = Unit.unit(unit) # I like this line.
        Unit._open_lib[alias] = unit
        unit._add_alias(alias)
        
    @staticmethod
    def load(library_name):
        import_module(library_name)
        
    @staticmethod
    def format_latex(text):
        if text == '':
            return ''
        text = text.replace(' ','\\,')
        if text.startswith('\t\t'):
            return text[2:]
        if text.startswith('\\'):
            return text
        if text.startswith('\t'):
            return '\t\\mathrm{' + text[1:] + '}'
        return '\\mathrm{' + text + '}'
              
    _gconv = None
    _conversion = None
    linear = True
    
    def __init__(self,name,symbol,conversion=None,short_name = None,
                 add_symbol=False,html_symbol=None,latex_symbol=None,
                 ascii_symbol=None,description=None,order = -1,**kwds):
        if conversion is not None and not isinstance(conversion,Conversion):
            raise ValueError('conversion is not an instance of the Conversion class')
        self.name = name
        self.conversion = conversion
        self.description=description
        self.order = order
        
        if symbol.startswith('\t\t'):
            raise ValueError('the symbol may not start with more than one tab character; only the latex_symbol may start with double tabs')
        self.symbol = symbol
        if html_symbol is None:
            self.html_symbol = symbol
        else:
            self.html_symbol=html_symbol
        if latex_symbol is None:
            self.latex_symbol = Unit.format_latex(symbol)
        else:
            self.latex_symbol = Unit.format_latex(latex_symbol)
        if ascii_symbol is None:
            self.ascii_symbol=symbol
        else:
            self.ascii_symbol=ascii_symbol
            
        self._aliases = set()
        
        self.parent = kwds.get('parent')
        
        #reset _used_units incase we are shadowing any unit definitions
        Unit._used_units = {} 
        
        Unit._open_lib[name] = self
        if add_symbol:
            if symbol.strip() != name:
                Unit._open_lib[symbol.strip()] = self
                self._aliases.add(symbol.strip())
            if (ascii_symbol is not None and ascii_symbol.strip() != name and 
                   ascii_symbol.strip() != symbol.strip()):
                Unit._open_lib[ascii_symbol.strip()] = self
                self._aliases.add(ascii_symbol.strip())
        if short_name is not None:
            Unit._open_lib[short_name] = self
            self.short_name = short_name
            self._aliases.add(short_name)
        else:
            if add_symbol:
                if ascii_symbol is not None:
                    self.short_name = ascii_symbol
                else:
                    self.short_name = symbol
            else:
                self.short_name = name

    @property
    def conversion(self):
        """
        Gets or sets the `Conversion` instance for the unit.
        
        ***This property is not intended to be used directly and setting this 
        property may cause problems***
        """
        return self._conversion
    @conversion.setter
    def conversion(self,c):
        self._conversion = c
        if c is not None:
            c.parent = self
            
    @property
    def aliases(self):
        """read-only
        
        Returns a set of the unshadowed aliases of this unit.  To add aliases
        to the unit use the `Unit.alias` static method.
        """
        return {a for a in self._aliases if (Unit.unit('[' + a + ']',exception=False) is self and a != self.name)}
    
    @property
    def shadowed_aliases(self):
        """read-only
        
        Returns a set of the shadowed aliases of this unit.
        """
        sha = {a for a in self._aliases if Unit.unit('[' + a + ']',exception=False) is not self}
        if Unit.unit('[' + self.name + ']') is not self:
            sha.add(self.name)
        return sha
    
    def _add_alias(self,alias):
        self._aliases.add(alias)
    
    def tostring(self,fmt=None,**kwds):
        """
        Returns a string containing the symbol for the unit the format given by
        the keyword `fmt` which may be set to a string the values 'html', 'latex',
        'ascii', or 'unicode'.  
        """
        if fmt is None or fmt == 'unicode':
            return self.symbol
        if fmt is None:
            return self.symbol
        if fmt == 'html':
            return self.html_symbol
        if fmt == 'latex':
            return self.latex_symbol
        if fmt == 'ascii':
            return self.ascii_symbol
        raise ValueError('format ' + str(fmt) + ' is not recognized')

    def _convs(self,d,c,h):          
        if c is None:
            c = self._conversion.copy()
        else:
            if self in h:
                if self is d:
                    raise CircularUnitConversionError('a circular conversion chain was found starting with unit ' + d.name + ' in library ' + d.library)
                else:
                    raise CircularUnitConversionError('a circular conversion chain was found;\nstarted search with unit ' + d.name + ' in library ' + d.library + ' and the loop started with unit ' + self.name + ' in library ' + self.library)
            h.add(self)
            d._gconv = c
            if self._conversion is None:
                return
            c = c.chain(self._conversion)
            
        try:
             u = self._conversion.unit
        except UnitNotFoundError:
            warnings.warn('while searching conversions:  unit ' + self.name + ' should have a conversion to ' + self._conversion.unit_name + ' in library ' + str(self._conversion.unit_library) + ' but the second unit was not found',UnitWarning,stacklevel=3)
            return
        except UnitLibNotFoundError:
            warnings.warn('while searching conversions:  unit ' + self.name + ' should have a conversion to ' + self._conversion.unit_name + ' in library ' + str(self._conversion.unit_library) + ' but the library containing the second unit has not been loaded',UnitWarning,stacklevel=3)
            return
        
        u._convs(d,c,h)
                
    def _convtree(self):
        if self._conversion is not None:
            self._convs(self,None,set([self]))
            
    def _convert(self,g,un):
        if self._conversion is not None and self._conversion.unit is un:
            return self._conversion.to(g)
            
        if self._gconv is not None and self._gconv._unit is un:
            return self._gconv.to(g)
            
        if un._conversion is not None and un._conversion.unit is self:
            return un._conversion.frm(g)
                
        if un._gconv is not None and un._gconv._unit is self:
            return un._gconv.frm(g)
        
        if self._gconv is not None and un._gconv is not None and un._gconv._unit is self._gconv._unit:
            g = self._gconv.to(g)
            g = un._gconv.frm(g)
            return g
            
        return None
        
    def convert(self,g,unit):
        """
        Converts a number or gummy from a quanitity with the units self to unit.
        
        This is not intended be used directly, instead use the `gummy.convert`
        method.
        """
        unit = Unit.unit(unit)
        if unit is self:
            return g
            
        ret = self._convert(g,unit)
        if ret is not None:
            return ret
        self._convtree()
        unit._convtree()
        ret = self._convert(g,unit)

        unm = unit.short_name
        if ret is None:
            raise NoUnitConversionFoundError('no conversion found from unit ' + self.short_name + ' to unit ' + unm)
        return ret
    
    @property
    def is_dimensionless(self):
        """
        `bool`, read-only
        
        Returns `True` if a conversion exists between `self` and `one`, and `False`
        if not.
        """
        if self.conversion is not None and self.conversion.unit is one:
            return True
        if self._gconv is not None and self._gconv._unit is one:
            return True
        return False
        
    @property
    def units(self):
        """read-only
        
        Returns a list of the constituent units and their exponents, e.g. for
        kg m**2/s units would be [(kg, 1), (m, 2), (s, -1)].
        """
        return [(self,1)]
    
    @staticmethod
    def _cancel(a,b=[],asign=1,bsign=1):
        # Takes .units lists a**asign and b**bsign and attempts to convert 
        # each unit in a into another unit either in a or b, and if successful 
        # cancels that unit.  Returns a tuple containing the  new .units 
        # list for the reduced composite unit a**asign * b**bsign as well as
        # the conversion factor between the unreduced and reduced composite unit.
        # In addition if both a an b contain dimensionless units then all
        # dimensionless units are converted to one.
        c = 1
        b = {i: bsign*j for i,j in b}
        for k,v in a:
            v *= asign
            if k in b:
                b[k] += v
            else:
                converted = False
                if not k.is_dimensionless:
                    for f in b.keys():
                        try:
                            c *= k.convert(1,f)**v
                            b[f] += v
                            converted = True
                            break
                        except NoUnitConversionFoundError:
                            pass
                if not converted:
                    b[k] = v
        un = []
        for k,v in b.items():
            if v != 0 and k is not one:
                uun,cc = k._cpow(v)
                un.append(uun[0])
                c *= cc
        return (un,c)
        
    def _mul_cancel(self,b):
        # self*b canceling units when possible
        c = Unit._cancel(b.units)
        d = Unit._cancel(self.units,c[0])
        return (_CompositeUnit(d[0]),c[1]*d[1])
        
    def _div_cancel(self,b):
        # self/b canceling units when possible        
        c = Unit._cancel(b.units)
        d = Unit._cancel(self.units,b=c[0],asign=1,bsign=-1)
        return (_CompositeUnit(d[0]),c[1]*d[1])
        
    def _rdiv_cancel(self,b):
        # b/self canceling units when possible
        c = Unit._cancel(b.units)
        d = Unit._cancel(self.units,b=c[0],asign=-1,bsign=1)
        return (_CompositeUnit(d[0]),c[1]*d[1])
    
    def _nummy_add(self,a,b):
        return nummy._add(a,b)

    def _add(self,a,bunit,b,aconv):
        if self is bunit:
            return (self._nummy_add(a,b),self)
        if not bunit.linear:
            return bunit._radd(b,self,a,not aconv)
        try:
            if aconv:
                return (self._nummy_add(self.convert(a,bunit),b),bunit)
            return (self._nummy_add(a,bunit.convert(b,self)),self)
        except NoUnitConversionFoundError:
            raise IncompatibleUnitsError('a quantity with unit ' + self.tostring() + ' may not be added to a quantity with unit ' + bunit.tostring())

    def _nummy_radd(self,a,b):
        # ummy._radd assumes that b is not an ummy.  
        if isinstance(b,nummy):
            return nummy._add(b,a)
        return nummy._radd(a,b)
    
    def _radd(self,a,bunit,b,aconv):
        if self is bunit:
            return (self._nummy_radd(a,b),self)
        try:
            if aconv:
                return (self._nummy_radd(self.convert(a,bunit),b),bunit)
            return (self._nummy_radd(a,bunit.convert(b,self)),self)
        except NoUnitConversionFoundError:
            raise IncompatibleUnitsError('a quantity with unit ' + self.tostring() + ' may not be added to a quantity with unit ' + bunit.tostring())
    
    def _nummy_sub(self,a,b):
        return nummy._sub(a,b)
    
    def _sub(self,a,bunit,b,aconv):
        if self is bunit:
            return (self._nummy_sub(a,b),self)
        try:
            if not bunit.linear:
                return bunit._rsub(b,self,a,not aconv)
            if aconv:
                return (self._nummy_sub(self.convert(a,bunit),b),bunit)
            return (self._nummy_sub(a,bunit.convert(b,self)),self)
        except NoUnitConversionFoundError:
            raise IncompatibleUnitsError('a quantity with unit ' + bunit.tostring() + ' may not be subtracted from a quantity with unit ' + self.tostring())

    def _nummy_rsub(self,a,b):
        if isinstance(b,nummy):
            return nummy._sub(b,a)
        return nummy._rsub(a,b)
    
    def _rsub(self,a,bunit,b,aconv):
        if self is bunit:
            return (self._nummy_rsub(a,b),self)
        try:
            if aconv:
                return (self._nummy_rsub(self.convert(a,bunit),b),bunit)
            return (self._nummy_rsub(a,bunit.convert(b,self)),self)
        except NoUnitConversionFoundError:
            raise IncompatibleUnitsError('a quantity with unit ' + self.tostring() + ' may not be subtracted from a quantity with unit ' + bunit.tostring())
        
    def _nummy_mod(self,a,b):
        return nummy._mod(a,b)
    
    def _mod(self,a,bunit,b,aconv):
        if self is bunit:
            return (self._nummy_mod(a,b),self)
        if not bunit.linear:
            return bunit._rmod(b,self,a,not aconv)
        try:
            if aconv:
                return (self._nummy_mod(self.convert(a,bunit),b),bunit)
            return (self._nummy_mod(a,bunit.convert(b,self)),self)
        except NoUnitConversionFoundError:
            raise IncompatibleUnitsError('a quantity with unit ' + self.tostring() + ' may not be added to a quantity with unit ' + bunit.tostring())

    def _nummy_rmod(self,a,b):
        if isinstance(b,nummy):
            return nummy._mod(b,a)
        return nummy._rmod(a,b)
    
    def _rmod(self,a,bunit,b,aconv):
        if self is bunit:
            return (self._nummy_rmod(a,b),self)
        try:
            if aconv:
                return (self._nummy_rmod(self.convert(a,bunit),b),bunit)
            return (self._nummy_rmod(a,bunit.convert(b,self)),self)
        except NoUnitConversionFoundError:
            raise IncompatibleUnitsError('a quantity with unit ' + self.tostring() + ' may not be added to a quantity with unit ' + bunit.tostring())

    def _nummy_mul(self,a,b):
        return nummy._mul(a,b)
    
    def _mul(self,a,bunit,b,aconv):
        if not bunit.linear:
            return bunit._rmul(b,self,a,not aconv)

        if aconv:
            un,c = self._mul_cancel(bunit)
        else:
            un,c = bunit._mul_cancel(self)
        
        return (self._nummy_mul(self._nummy_mul(a,b),c),un)
    
    def _nummy_rmul(self,a,b):
        if isinstance(b,nummy):
            return nummy._mul(b,a)
        return nummy._rmul(a,b)

    def _rmul(self,a,bunit,b,aconv):
        if aconv:
            un,c = self._mul_cancel(bunit)
        else:
            un,c = bunit._mul_cancel(self)
        
        return (self._nummy_mul(self._nummy_rmul(a,b),c),un)
    
    def _nummy_truediv(self,a,b):
        return nummy._truediv(a,b)
    
    def _truediv(self,a,bunit,b,aconv):
        if not bunit.linear:
            return bunit._rtruediv(b,self,a,not aconv)

        if aconv:
            un,c = self._div_cancel(bunit)
        else:
            un,c = bunit._rdiv_cancel(self)
        
        return (self._nummy_mul(self._nummy_truediv(a,b),c),un)
    
    def _nummy_rtruediv(self,a,b):
        if isinstance(b,nummy):
            return nummy._truediv(b,a)
        return nummy._rtruediv(a,b)

    def _rtruediv(self,a,bunit,b,aconv):
        if aconv:
            un,c = self._rdiv_cancel(bunit)
        else:
            un,c = bunit._div_cancel(self)
        
        return (self._nummy_mul(self._nummy_rtruediv(a,b),c),un)
    
    def _cpow(self,v):
        return ([[self,v]],1)
    
    def _nummy_pow(self,a,b):
        return nummy._pow(a,b)
    
    def _pow(self,a,bunit,b,aconv):
        if not bunit.linear:
            return bunit._rpow(b,self,a,not aconv)
        
        if bunit is not one:
            try:
                b = bunit.convert(b,one)
            except NoUnitConversionFoundError:
                raise IncompatibleUnitsError('the exponent is not dimensionless')
                
        bb = float(b)
        if bb != b:
            try:
                a = self.convert(a,one)
            except NoUnitConversionFoundError:
                raise IncompatibleUnitsError('a gummy that is not dimensonless may not be raised to a power with an uncertainty')
            return (self._nummy_pow(a,b),one)
        b = float(b)
        if int(b) == b:
            b = int(b)
        
        if b == 0:
            return (self._nummy_pow(a,0),one)
        
        un,c = self._cpow(b)
        un = _CompositeUnit(un)
        
        return (self._nummy_mul(self._nummy_pow(a,b),c),un)

    def _nummy_rpow(self,a,b):
        if isinstance(b,nummy):
            return nummy._pow(b,a)
        return nummy._rpow(a,b)
    
    def _rpow(self,a,bunit,b,aconv):
        if self is not one:
            try:
                b = bunit.convert(b,one)
            except NoUnitConversionFoundError:
                raise IncompatibleUnitsError('the exponent is not dimensionless')
                
        aa = float(a)
        if aa != a:
            try:
                b = self.convert(b,one)
            except NoUnitConversionFoundError:
                raise IncompatibleUnitsError('a gummy that is not dimensonless may not be raised to a power with an uncertainty')
            return (self._nummy_rpow(a,b),one)
        a = float(a)
        if int(a) == a:
            a = int(a)
        
        if a == 0:
            return (self._nummy_pow(b,0),one)
        
        un,c = _CompositeUnit(bunit._cpow(a))
        
        return (self._nummy_mul(self._nummy_rpow(a,b),c),un)
    
    def _nummy_neg(self,a):
        return nummy.__neg__(a)
    
    def _neg(self,a):
        return (self._nummy_neg(a),self)
    
    def _nummy_pos(self,a):
        return nummy.__pos__(a)
    
    def _pos(self,a):
        return (self._nummy_pos(a),self)
    
    def _nummy_abs(self,a):
        return nummy.__abs__(a)
    
    def _abs(self,a):
        return (self._nummy_abs(a),self)
    
    def __mul__(self,v):          
        if v is one:
            return self
        
        if not isinstance(v,Unit):
            try:
                if v == 1:
                    return self
            except:
                pass
            raise TypeError('unsupported operand type(s) for *: Unit and ' + str(v))
            
        if not v.linear:
            return v.__rmul__(self)
            
        return _CompositeUnit(self.units + v.units)
        
    def __truediv__(self,v):
        if v is one:
            return self
        
        if not isinstance(v,Unit):
            try:
                if v == 1:
                    return self
            except:
                pass
            raise TypeError('unsupported operand type(s) for /: Unit and ' + str(v))
            
        if not v.linear:
            return v.__rtruediv__(self)
            
        vi = [(e[0],-e[1]) for e in v.units]
        return _CompositeUnit(self.units + vi)

    def __rtruediv__(self,v):
        if v is one or v == 1:
            return _CompositeUnit([(e[0],-e[1]) for e in self.units])
        
    def __pow__(self,v):
        try:
            if v == 0:
                return one
            
            if v == 1:
                return self
        except:
            raise TypeError('unsupported operand type(s) for **: Unit and ' + str(v))
        
        if v != float(v):
            raise TypeError('unsupported operand type(s) for **: Unit and ' + str(v))

        return _CompositeUnit([(e[0],v*e[1]) for e in self.units])
        
    # Define comparison operators and a hash function so, for the 
    # _CompositeUnit subclass we can sort the Unit order and store them as keys 
    # in a unique OrderedDict.
    def __lt__(self, b):
        if self.order == b.order:
            return self.name < b.name
        return self.order < b.order

    def __le__(self, b):
        if self.__eq__(b):
            return True
        if self.order == b.order:
            return self.name < b.name
        return self.order < b.order

    def __eq__(self, b):
       return (self is b)

    def __ge__(self, b):
        if self.__eq__(b):
            return True
        if self.order == b.order:
            return self.name > b.name
        return self.order > b.order

    def __gt__(self, b):
        if self.order == b.order:
            return self.name > b.name
        return self.order > b.order

    def __ne__(self, b):
        return not (self is b)
        
    def __hash__(self):
        return id(self)
        
    
class _CompositeUnit(Unit):
    # Instances of this class represent composite derived unit.
    
    living_units = weakref.WeakValueDictionary()
    
    parent = None
    
    @staticmethod
    def _unicode_super(e):
        if np.modf(e)[0] == 0:
            se = str(int(e))
            so = ''
            for s in se:
                if s == '-':
                    so += '\u207B'
                if s == '0':
                    so += '\u2070'
                if s == '1':
                    so += '\u00B9'
                if s == '2':
                    so += '\u00B2'
                if s == '3':
                    so += '\u00B3'
                if s == '4':
                    so += '\u2074'
                if s == '5':
                    so += '\u2075'
                if s == '6':
                    so += '\u2076'
                if s == '7':
                    so += '\u2077'
                if s == '8':
                    so += '\u2078'
                if s == '9':
                    so += '\u2079'
            return so
        return '**' + '{:.2f}'.format(e).rstrip('0')
    
    @staticmethod
    def _format_sscript(e):
        mf = np.modf(e)
        if mf[0] == 0:
            return '{:.0f}'.format(e)
        if mf[0] == 0.5:
            return '{:.0f}'.format(2*e) + '/2'
        if mf[0] == 0.25:
            return '{:.0f}'.format(4*e) + '/4'
        if np.modf(3*e)[0] < 0.01:
            return '{:.0f}'.format(3*e) + '/3'
        return '{:.2f}'.format(e).rstrip('0')
        
    @staticmethod
    def _iaddsym(txt,s,e,fmt,f):
        if txt.startswith('\t'):
            txt = txt[1:]
        if e == 0:
            return txt
        if f == 1 and e < 0:
            return txt
        if f == 2:
            if e > 0:
                return txt
            e = -e
            
        if fmt == 'html':
            if txt != '':
                txt += '&middot;'
            txt += s
            if e != 1:
                txt += '<sup>' + _CompositeUnit._format_sscript(e) + '</sup>'
        elif fmt == 'latex':
            if txt != '':
                txt += r'\cdot'
            txt += s
            if e != 1:
                txt += '^{' + _CompositeUnit._format_sscript(e) + '}'
        elif fmt == 'ascii':
            if txt != '':
                txt += '*'
            txt += s
            if e != 1:
                txt += '**' + '{:.2f}'.format(e).rstrip('0').rstrip('.')
        else:
            if txt != '':
                txt += ' '
            txt += s
            if e != 1:
                txt += _CompositeUnit._unicode_super(e)
            
        return txt
        
    def _addsym(txt,u,fmt=None,f=0):
        return _CompositeUnit._iaddsym(txt,u[0].tostring(fmt=fmt),u[1],fmt,f)
    
    def __new__(cls,ul):
        ul = list(ul)
        
        if len(ul) == 0:
            return one
        
        if len(ul) == 1 and ul[0][1] == 1:
            return ul[0][0]
        
        ul = sorted(ul,key = lambda x: (x[0],-x[1]))
        
        uid = tuple([(id(u[0]),u[1]) for u in ul])
        ret = _CompositeUnit.living_units.get(uid)
        if ret is not None:
            return ret
        
        linear = True
        for u,e in ul:
            if not u.linear:
                linear = False
                try:
                    cls = u.get_composite(ul)
                except NotImplementedError:
                    raise IncompatibleUnitsError('unit ' + u.tostring() + ' may not be combined with other units')
                break
        
        ret = object.__new__(cls)
        
        ret.linear = linear
        ret.order = 100
        ret._units = ul
        _CompositeUnit.living_units[uid] = ret
        ret._oconv = None
        ret._have_conversion = False
        ret._aliases = set()
        
        nm = ''
        snm = ''
        for u in ul:          
            nm += ' '
            snm += ' '
            nmi = u[0].name
            if ' ' in nmi and (not nmi[0].startswith('[') or not nmi.endswith(']')):
                nmi = '[' + nmi + ']'
            snmi = u[0].short_name
            if ' ' in snmi and (not snmi[0].startswith('[') or not snmi.endswith(']')):
                snmi = '[' + snmi + ']'
            if u[1] == 1:
                    nm += nmi
                    snm += snmi
            else:
                    nm += nmi + '**' + str(u[1])
                    snm += snmi + '**' + str(u[1])
        
        ret.name = nm.strip()
        ret.short_name = snm.strip()
        
        ret._make_symbols(ul)
        
        return ret
    
    def __init__(self,ul):
        pass
        
    def _make_symbols(self,ul):
        sym = ''
        hsym = ''
        lsym = ''
        asym = ''
        symsn = ''
        symsd = ''
        hsymsn = ''
        hsymsd = ''
        lsymsn = ''
        lsymsd = ''
        asymsn = ''
        asymsd = ''
        
        nd = 0
        for u in ul:
            sym = _CompositeUnit._addsym(sym,u)
            hsym = _CompositeUnit._addsym(hsym,u,fmt = 'html')
            lsym = _CompositeUnit._addsym(lsym,u,fmt = 'latex')
            asym = _CompositeUnit._addsym(asym,u,fmt = 'ascii')
            
            symsn = _CompositeUnit._addsym(symsn,u,f=1)
            hsymsn = _CompositeUnit._addsym(hsymsn,u,fmt = 'html',f=1)
            lsymsn = _CompositeUnit._addsym(lsymsn,u,fmt = 'latex',f=1)
            asymsn = _CompositeUnit._addsym(asymsn,u,fmt = 'ascii',f=1)
            
            symsd = _CompositeUnit._addsym(symsd,u,f=2)
            hsymsd = _CompositeUnit._addsym(hsymsd,u,fmt = 'html',f=2)
            lsymsd = _CompositeUnit._addsym(lsymsd,u,fmt = 'latex',f=2)
            asymsd = _CompositeUnit._addsym(asymsd,u,fmt = 'ascii',f=2)
            
            if u[1] < 0:
                nd += 1
            
        self.sym = sym
        self.hsym = hsym
        self.lsym = lsym
        self.asym = asym
        
        if symsn == '' and nd > 0:
            symsn = '1'
            hsymsn = '1'
            lsymsn = '1'
            asymsn = '1'
        
        if nd > 0:
            if nd > 1:
                symsd = '(' + symsd + ')'
                hsymsd = '(' + hsymsd + ')'
                lsymsd = '(' + lsymsd + ')'
                asymsd = '(' + asymsd + ')'
            syms = symsn + '/' + symsd
            hsyms = hsymsn + '/' + hsymsd
            lsyms = lsymsn + '/' + lsymsd
            asyms = asymsn + '/' + asymsd
        else:
            syms = symsn
            hsyms = hsymsn
            lsyms = lsymsn
            asyms = asymsn
        self.syms = syms
        self.hsyms = hsyms
        self.lsyms = lsyms
        self.asyms = asyms
        
    @property
    def units(self):
        """
        read-only

        Returns a list of the constituent units and their exponents, e.g. for
        kg m**2/s units would be [(kg, 1), (m, 2), (s, -1)].
        """
        return self._units
        
    def tostring(self,fmt=None,solidus=True,mulsep=False,**kwds):
        """
        Returns a string containing the symbol for the unit the format given by
        the keyword fmt which may be set to a string the values 'html', 'latex', 
        'ascii', or 'unicode'.
        
        If solidus is `True` a slash will be used instead of negative exponents
        and if `mulsep` is `True` a mid-dot or asterix will be used instead of a space
        to separate units.
        """
        if fmt is None or fmt == 'utf-8' or fmt == 'unicode':
            if solidus:
                ret = self.syms
            else:
                ret = self.sym
            if not mulsep:
                ret = ret.replace('\u00B7',' ')
            return ret
        elif fmt == 'html':
            if solidus:
                ret = self.hsyms
            else:
                ret = self.hsym
            if not mulsep:
                ret = ret.replace('&middot;','&thinsp;')
        elif fmt == 'latex':
            if solidus:
                ret = self.lsyms
            else:
                ret = self.lsym
            if not mulsep:
                ret = ret.replace(r'\cdot',r'\,')
            return ret
        elif fmt == 'ascii':
            if solidus:
                ret = self.asyms
            else:
                ret = self.asym
            if not mulsep:
                ret = ret.replace('**','#@@#')
                ret = ret.replace('*',' ')
                ret = ret.replace('#@@#','**')
        else:
            raise ValueError('format ' + str(fmt) + ' is unrecognized')
        return ret
        
    def _find_conv(self):    
        c = None
        cun = {}
        cg = None
        cgun = {}
        for un,e in self._units:
            if un.conversion is None:
                if un in cun:
                    cun[un] += e
                else:
                    cun[un] = e
            else:
                p = un.conversion.pow(e)
                if c is None:
                    c = p
                else:
                    c = c.chain(p)
                for t in p.unit.units:
                    if t[0] in cun:
                        cun[t[0]] += t[1]
                    else:
                        cun[t[0]] = t[1]
                    
            un._convtree()
                    
            if un._gconv is None:
                if un in cgun:
                    cgun[un] += e
                else:
                    cgun[un] = e
            else:
                p = un._gconv.pow(e)
                if cg is None:
                    cg = p
                else:
                    cg = c.chain(p)
                for t in p.unit.units:
                    if t[0] in cgun:
                        cgun[t[0]] += t[1]
                    else:
                        cgun[t[0]] = t[1]
                        
        cun = [(k,v) for k,v in cun.items() if v != 0]
        cgun = [(k,v) for k,v in cgun.items() if v != 0]
        
        cun = _CompositeUnit(cun)
        if cun is not self:
            if c is None:
                c = Conversion(cun,1)
            else:
                c._unit = cun
            c.parent = self
            self._conversion = c
        else:
            self._conversion = None
        
        cgun = _CompositeUnit(cgun)
        if cgun is not self:
            if cg is None:
                cg = Conversion(cgun,1)
            else:
                cg._unit = cgun
            cg.parent = self
            self._gconv = cg
        else:
            self._gconv = None

        self._have_conversion = True

    
    def _cpow(self,v):
        c = 1
        un = []
        for u,e in self.units:
            uun,cc = u._cpow(v*e)
            un.append(uun[0])
            c *= cc
        return (un,c)

    @property
    def conversion(self):
        """
        Gets or sets the conversion object for the unit.  
        
        ***This property is not intended to be used directly and setting this 
        property may cause problems***
        """
        if not self._have_conversion:
            self._find_conv()
        return self._conversion
    
    @property
    def shadowed_aliases(self):
        """
        Returns a set of the shadowed aliases of this unit.
        """
        sha = {a for a in self._aliases if Unit.unit('[' + a + ']',exception=False) is not self}
        return sha
        
    def _convtree(self):
        if not self._have_conversion:
            self._find_conv()
        Unit._convtree(self)
        
    def _convs(self,d,c,h):
        if not self._have_conversion:
            self._find_conv()
        Unit._convs(self,d,c,h)

    
class _BuiltinLib:
    """
    Allows Unit objects to be added to Unit._builtin_lib rather than Unit._lib
    using a "with" statement.
    
    Example
    -------
    
    >>> with _BuiltinLib():
    ...     one = Unit('1','',add_symbol=False)
    
    """
    def  __enter__(self):
        Unit._open_lib = Unit._builtin_lib
    def __exit__(self,type, value, traceback):
        Unit._open_lib = Unit._lib
    
class _One(Unit):
    """
    The only instance of this class should be the unit `one`.
    """
    def __mul__(self,v):            
        return v
        
    def __truediv__(self,v):
        return 1/v

    def __rtruediv__(self,v):
        return v
        
    def __pow__(self,v):
        return self
    
    @property
    def is_dimensionless(self):
        return True
    
with _BuiltinLib():
    one = _One('1','',add_symbol=False)


# -*- coding: utf-8 -*-

import metrolopy as uc
import numpy as np
from fractions import Fraction

try:
    import mpmath as mp
except:
    mp = None

rand = np.random.RandomState()

def display(g):
    w = rand.randint(5)
    assert '?' not in g.tostring()
    if w == 0:
        print(g)
    elif w == 1:
        g.latex()
    elif w == 2:
        g.html()
    elif w == 3:
        g.unicode()
    else:
        g.ascii()

def make_gummy(sign=None,exp=None,uexp=-6,sometimes_small=True,dof=None,
               unit=None,units=None,mpf=False,allowzero=True,allowlargeu=True):
    
    utypes = ['A','B','D',None]
    if units is None:
        units = ['m','lb','m**2 s**3/kg**4','m**3','s**-2','1']
        
    if rand.randint(10) == 0:
        if exp is None:
            x = rand.randint(6000)
        else:
            x = rand.randint(int(10**exp))
    else:
        x = rand.rand()*0.9 + 0.1
        
        if exp is None:
            exp = rand.randint(1,6)
            if rand.randint(2):
                exp = -exp
                
        x = x*10**exp
    
    if sign is None:
        if rand.randint(2):
            x = -x
    else:
        x *= sign
        
    if not allowzero and x == 0:
        x = x + 1
        
    if allowlargeu and rand.randint(10) == 0:
        u = rand.randint(20) + 1
    else:
        u = rand.rand()*0.5 + 0.5
        if sometimes_small and not rand.randint(10):
            uexp -= 4
        u = abs(x)*u*10**uexp
    
    if mpf and mp is not None and not rand.randint(4):
        x = mp.mpf(x)
        if rand.randint(2):
            u = mp.mpf(u)
        
    if dof is None:
        if not rand.randint(4):
            dof = float('inf')
        else:
            dof = rand.randint(5,15)
            
    if unit is None:
        unit = units[rand.randint(len(units))]

    utype = utypes[rand.randint(len(utypes))]
    
    g = uc.gummy(x,u=u,dof=dof,unit=unit,utype=utype)
    
    return (g,x,u,dof,unit,utype)

def make_number(sign=None,gconst=True,exp=None,fionly=True,allowzero=True):
    q = rand.randint(4)
    if fionly and q > 1:
        q = 1
    if q == 0:
        if exp is not None:
            x = rand.randint(10**(exp+1))
        else:
            x = rand.randint(100000)
    if q == 1 or q == 2:
        x = rand.rand()*0.9 + 0.1
        if exp is None:
            exp = rand.randint(1,6)
            if rand.randint(2):
                exp = -exp

        x = x*10**exp
    if q == 2 and mp is not None:
        x = mp.mpf(x)
    elif q == 3:
        if exp is not None:
            a = rand.randint(10**(exp+1))
            b = rand.randint(10**(exp+1))
        else:
            a = rand.randint(100000)
            b = rand.randint(100000)
        if b == 0:
            x = Fraction(0)
        else:
            x = Fraction(a,b)
        
    
    if sign is None:
        if rand.randint(2):
            x = -x
    else:
        x *= sign
    
    if gconst and rand.randint(2):
        x = uc.gummy(x)
    
    if not allowzero and x == 0:
        x = x + 1
        
    return x


# -*- coding: utf-8 -*-

import metrolopy as uc
import numpy as np

from metrolopy.tests.common import rand,make_gummy

def test_ubreakdown(n=None,prnt=False,budget=False):
    uc.gummy.p_method = None
    
    if n is None:
        if prnt:
            n = 100
        else:
            n = 10000
        
    for i in range(n):
        k = rand.randint(1,6)
        
        if rand.randint(2):
            unt = '1'
        else:
            unt = None
        g,x,u,dof,unit,utype = zip(*[make_gummy(unit=unt) for y in range(k)])
        
        for i,y in enumerate(g):
            if y._u != 0:
                assert y.utype is utype[i]
        
        c = rand.rand()*0.9 + 0.1
        if rand.randint(2):
            c = -c
        
        def f(*p):
            r = c
            for y in p:
                r *= y
            return r
        
        def d(k,*p):
            r = c
            for i,y in enumerate(p):
                if i != k:
                    r *= y
            return r
        
        gr = f(*g)
        
        ut = None
        for q in g:
            if gr._ref is q._ref:
                ut = q.utype
        assert gr.utype is ut
        
        x = list(x)
        u = list(u)
        lb = 'lb' in str(gr.unit)
        kg = 'kg' in str(gr.unit)
        for i,un in enumerate(unit):
            if 'lb' in un and not lb:
                x[i] *= 0.45359237
                u[i] *= 0.45359237
            if 'kg' in un and not kg and lb:
                x[i] *= 0.45359237**4
                u[i] *= 0.45359237**4
                
        xr = f(*x)
        
        if xr == 0:
            assert abs(gr.x - xr) < 1e-10
        else:
            assert abs(gr.x - xr)/abs(xr) < 1e-10
        
        ur = 0
        for i,y in enumerate(u):
            ur += y**2*d(i,*x)**2
        ur = np.sqrt(ur)
        if ur == 0:
            assert abs(gr.u) < 1e-10
        else:
            assert abs(gr.u - ur)/ur < 1e-10
        
        dofr = 0
        for i,(v,idof) in enumerate(zip(u,dof)):
            dofr += v**4*d(i,*x)**4/idof
        if dofr == 0:
            dofr = float('inf')
        else:
            dofr = ur**4/dofr
        if dofr > uc.gummy.max_dof:
            dofr = float('inf')
            
        # Due to the fact that any dof larger than gummy.mox_dof is rounded to 
        # infinity in intermediate calculations, gr.dof may not be exactly dofr
        # particularly for large dofr.  
        if dofr > 100:
            assert gr.dof > 100
        else:
            assert abs(gr.dof - dofr)/dofr < 1e-2
            
        if unt == '1':
            if k == 1:
                dr = lambda *x: d(0,*x)
            else:
                dr = lambda *x: [d(i,*x) for i in range(k)]
            ga = uc.gummy.apply(f,dr,*g)
            if xr == 0:
                assert abs(ga.x) < 1e-10
            else:
                assert abs(ga.x - xr)/abs(xr) < 1e-10
            if ur == 0:
                assert abs(ga.u) < 1e-6
            else:
                assert abs(ga.u - ur)/ur < 1e-6
            if dofr > 100:
                assert ga.dof > 100
            else:
                assert abs(ga.dof - dofr)/dofr < 1e-2
                
            gn = uc.gummy.napply(f,*g)
            if xr == 0:
                assert abs(gn.x) < 1e-10
            else:
                assert abs(gn.x - xr)/abs(xr) < 1e-10
            if ur == 0:
                assert abs(gn.u) < 1e-3
            else:
                assert abs(gn.u - ur)/ur < 1e-3
            if dofr > 100:
                assert gn.dof > 100
            else:
                assert abs(gn.dof - dofr)/dofr < 1e-2
                
            if rand.randint(2):
                if rand.randint(2):
                    gf = ga
                else:
                    gf = gn
            else:
                gf = gr
        else:
            gf = gr
            
        gl = list(g)
        if not rand.randint(4):
            gl.append(make_gummy(unit=unt)[0])
            
        if rand.randint(2):
            e = gl[rand.randint(len(gl))]
        else:
            e = [gl[rand.randint(len(gl))] for i in range(rand.randint(len(gl)))]
            
        if ur > 0:
            ua = 0
            ub = 0
            ud = 0
            ue = 0
            dofa = 0
            dofb = 0
            dofd = 0
            dofe = 0
            for i,(ig,iu,idof) in enumerate(zip(g,u,dof)):
                if gf.ufrom(ig) == 0:
                    assert iu*d(i,*x)/ur < 1e-3
                else:
                    if ig.utype == 'A':
                        ua += iu**2*d(i,*x)**2
                        dofa += iu**4*d(i,*x)**4/idof
                    elif ig.utype == 'B':
                        ub += iu**2*d(i,*x)**2
                        dofb += iu**4*d(i,*x)**4/idof
                    elif ig.utype == 'D':
                        ud += iu**2*d(i,*x)**2
                        dofd += iu**4*d(i,*x)**4/idof
                    
                    try:
                        if e is ig or ig in e:
                            ue += iu**2*d(i,*x)**2
                            dofe += iu**4*d(i,*x)**4/idof
                    except TypeError:
                        pass
                
            ua = np.sqrt(ua)
            ub = np.sqrt(ub)
            ud = np.sqrt(ud)
            ue = np.sqrt(ue)
            
            if dofa == 0:
                dofa = float('inf')
            else:
                dofa  = ua**4/dofa
            if dofb == 0:
                dofb = float('inf')
            else:
                dofb  = ub**4/dofb
            if dofd == 0:
                dofd = float('inf')
            else:
                dofd  = ud**4/dofd
            if dofe == 0:
                dofe = float('inf')
            else:
                dofe  = ue**4/dofe
                        
            if ua == 0:
                assert gf.ufrom('A') == 0
            else:
                assert abs(gf.ufrom('A') - ua)/ur < 1e-3
            if ua/ur > 0.01:
                if dofa > 100 or gf.ufrom('A') == 0:
                    assert gf.doffrom('A') > 100
                else:
                    assert abs(gf.doffrom('A') - dofa)/dofa < 1e-2
                
            if ub == 0:
                assert gf.ufrom('B') == 0
            else:
                assert abs(gf.ufrom('B') - ub)/ur < 1e-3
            if ub/ur > 0.01:
                if dofb > 100 or gf.ufrom('B') == 0:
                    assert gf.doffrom('B') > 100
                else:
                    assert abs(gf.doffrom('B') - dofb)/dofb < 1e-2
                
            if ud == 0:
                assert gf.ufrom('D') == 0
            else:
                assert abs(gf.ufrom('D') - ud)/ur < 1e-3
            if ud/ur > 0.01:
                if dofd > 100 or gf.ufrom('D') == 0:
                    assert gf.doffrom('D') > 100
                else:
                    assert abs(gf.doffrom('D') - dofd)/dofd < 1e-2
                
            if ue == 0:
                assert gf.ufrom(e) == 0
            else:
                assert abs(gf.ufrom(e) - ue)/ur < 1e-3
            #if ue/ur > 0.01:
                #if dofe > 100 or gf.ufrom(e) == 0:
                    #assert gf.doffrom(e) > 100
                #else:
                    #assert abs(gf.doffrom(e) - dofe)/dofe < 1e-2
                    
        if not rand.randint(4):
            gf.uunit = '%'
                
        styles = ['pm','pmi','concise','ueq','u','uf']
        gf.style = styles[rand.randint(6)]
                
        assert '?' not in gf.tostring()
        
        ub = {i.utype for i in g if i.utype is not None}
        if not rand.randint(8) and len(ub) > 1:
            ub.pop()
        ub = list(ub)
        
        gfub = gf.copy(formatting=True)
        if len(ub) > 0:
            gfub.ubreakdown = ub
        
        assert '?' not in gfub.tostring()
                
        if prnt:
            w = rand.randint(5)
            if w == 0:
                print(gf)
                print(gfub)
            elif w == 1:
                gf.latex()
                gfub.latex()
            elif w == 2:
                gf.html()
                gfub.html()
            elif w == 3:
                gf.unicode()
                gfub.unicode()
            else:
                gf.ascii()
                gfub.ascii()
        
        if budget:
            show_expanded_u = bool(rand.randint(2))
            show_subtotals = bool(rand.randint(2))
            
            k = None
            p = None
            if bool(rand.randint(2)):
                if bool(rand.randint(2)):
                    if bool(rand.randint(2)):
                        gf.k = 2
                    else:
                        k = 2
                else:
                    if bool(rand.randint(2)):
                        gf.p = 0.95
                    else:
                        p = 0.95
                        
            if bool(rand.randint(2)):
                gl[0].name = 'gl0'
                
            if bool(rand.randint(2)) and len(gl) > 1:
                gl[1].name = 'gl1'
                
            if bool(rand.randint(2)):
                gf.name = 'gf'
            
            if not bool(rand.randint(4)):
                description = ['description ' + str(i) for i in range(len(gl)+1)]
            else:
                description = None
                
            if not bool(rand.randint(4)):
                custom = [str(i) for i in range(len(gl)+1)]
                custom_heading = 'n'
            else:
                custom = None
                custom_heading = None
                
            show_s = None
            show_c = None
            show_d = None
            if bool(rand.randint(2)):
                show_s = bool(rand.randint(2))
                show_c = bool(rand.randint(2))
                show_d = bool(rand.randint(2))
                
            b = gf.budget(gl,show_subtotals=show_subtotals,
                          show_expanded_u=show_expanded_u,k=k,p=p,
                          description=description,show_s=show_s,show_c=show_c,
                          show_d=show_d,custom=custom,
                          custom_heading=custom_heading)
            
            w = rand.randint(5)
            if w == 0:
                print(b)
            elif w == 1:
                b.latex()
            elif w == 2:
                b.html()
            else:
                b.unicode()
        
        if prnt or budget:
            print()
            print('---')
            print()




# -*- coding: utf-8 -*-

# module offsetunit

# Copyright (C) 2019 National Research Council Canada
# Author:  Harold Parks

# This file is part of MetroloPy.

# MetroloPy is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.

# MetroloPy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more 
# details.

# You should have received a copy of the GNU General Public License along with 
# MetroloPy. If not, see <http://www.gnu.org/licenses/>.

"""
The OffsetUnit class was created to handle Celsius and Fahrenheit units.
"""

from .unit import Unit,Conversion
from .nonlinearunit import NonlinearUnit,NonlinearConversion
from .exceptions import IncompatibleUnitsError

class OffsetConversion(NonlinearConversion):
    """
    Represents a conversion of the form x -> x + offset.
    """
    def __init__(self,unit,offset):
        self._unit = unit
        self.offset = offset
        
    def _to(self,g):
        return g + self.offset
    
    def _frm(self,g):
        return g - self.offset
        
    def copy(self):
        r = OffsetConversion(self._unit,self.offset)
        r.parent = self.parent
        return r
        

class _IntervalUnit(Unit):
    
    def __init__(self,parent,*params,**kwds):
        conv = Conversion(parent.conversion.unit,1)
        if 'short_name' in kwds:
            kwds['short_name'] += '-i'
        elif kwds.get('add_symbol'):
            if 'ascii_symbol' in kwds:
                kwds['short_name'] = kwds['ascii_symbol'] + '-i'
            else:
                kwds['short_name'] = params[1] + '-i'
        kwds['add_symbol'] = False
        
        kwds['parent'] = parent
        Unit.__init__(self,params[0] + ' interval',params[1],conversion=conv,**kwds)
    
    def _getme(self,ul,e):
        return self.parent
    
class OffsetUnit(NonlinearUnit):
    """
    This class was created to handle units such as the degree Celsius and the
    degree Fahrenheit.  This class takes the same parameters as the Unit class,
    but actually creates two unit instances...
    """
    def __init__(self,*params,**kwds):            
        Unit.__init__(self,*params,**kwds)
        
        self.interval_unit = _IntervalUnit(self,*params,**kwds)
        
    def zero(self):
        return -self.conversion.offset
        
    def _add_alias(self,alias):
        self._aliases.add(alias)
        Unit.alias(alias + '-i',self.interval_unit)
        
    def _add(self,a,bunit,b,aconv):
        if bunit is not self.interval_unit:
            raise IncompatibleUnitsError('a quantity with unit ' + self.tostring() + ' may only be added to its interval unit counterpart')
        return (self._nummy_add(a,b),self)
    
    def _radd(self,a,bunit,b,aconv):
        if bunit is not self.interval_unit:
            raise IncompatibleUnitsError('a quantity with unit ' + self.tostring() + ' may only be added to its interval unit counterpart')
        return (self._nummy_radd(a,b),self)
    
    def _sub(self,a,bunit,b,aconv):
        if bunit is self:
            return (self._nummy_sub(a,b),self.interval_unit)
        if bunit is self.interval_unit:
            return (self._nummy_sub(a,b),self)
        raise IncompatibleUnitsError('a quantity with unit ' + bunit.tostring() + ' may not be subtracted from a quantity with unit ' + self.tostring() + '; automatic conversion is disabled with offset unit instances')
    
    def _rsub(self,a,bunit,b,aconv):
        if bunit is self:
            return (self._nummy_rsub(a,b),self.interval_unit)
        raise IncompatibleUnitsError('a quantity with unit ' + self.tostring() + ' may not be subtracted from a quantity with unit ' + bunit.tostring() + '; automatic conversion is disabled with offset unit instances')

        


# -*- coding: utf-8 -*-

# module ummy

# Copyright (C) 2019 National Research Council Canada
# Author:  Harold Parks

# This file is part of MetroloPy.

# MetroloPy is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.

# MetroloPy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more 
# details.

# You should have received a copy of the GNU General Public License along with 
# MetroloPy. If not, see <http://www.gnu.org/licenses/>.

"""
The ummy object defined here was created primarily to separate the code for
first-order uncertainty propagation from the other methods in the gummy object.  
But the ummy object can be used as a stripped down verison of gummy without the
Monte-Carlo uncertainty propagation, unit, and pretty printing functionality
of gummy.
"""

import numpy as np
import weakref
from .dfunc import Dfunc
from .exceptions import UncertiantyPrecisionWarning
from collections import namedtuple
from warnings import warn
from math import isnan,isinf,log,log10,pi
from fractions import Fraction
from numbers import Rational,Integral

try:
    from mpmath import mp,mpf,rational
except:
    mp = mpf = None

_FInfo = namedtuple('_FIinfo','rel_u precision warn_u ')
_iinfo = _FInfo(rel_u=0,precision=0,warn_u=0)
_finfodict = {}
def _getfinfo(x):
    if isinstance(x,Rational):
        return (_iinfo,x)
    
    if mpf is not None and isinstance(x,mpf):
        n = mp.prec
        x = mpf(x)
        try:
            return (_finfodict[n],x)
        except:
            f = _FInfo(rel_u=0.4222/2**n,
                       precision=mp.dps,
                       warn_u = 0)
            _finfodict[n] = f
            return (f,x)
        
    t = type(x)
    try:
        return (_finfodict[t],x)
    except:
        try:
            f = np.finfo(t)
            f = _FInfo(rel_u=f.eps*0.2111,
                       precision=f.precision,
                       warn_u=100*f.tiny*f.eps)
            _finfodict[t] = f
        except:
            _finfodict[t] = _iinfo
            return _iinfo
        return (f,x)

def _combu(a,b,c):
    # function for combining uncertainties,
    # computes sqrt(a**2 + 2*c*a*b + b**2)
    # with 1 >= c >= -1,
    # this method avoids overflows or underflows when taking the squares
    
    if isnan(a) or isnan(b):
        try:
            return type(a)('nan')
        except:
            return float('nan')
    
    if abs(b) > abs(a):
        a,b = b,a
    if b == 0:
        return abs(a)
        
    r = b/a
    x = 1 + r**2 + 2*c*r
    
    if x < 0:
        if x < -1e6:
            # something horrible happened, should only get here due to a bug
            raise FloatingPointError('u is sqrt of ' + str(a**2*x))
        # maybe possibly get here due to a floating point rounding error somewhere
        return 0
    
    return abs(a)*x**0.5     

def _isscalar(x):
    try:
        len(x)
        return False
    except:
        return True
    
def _floor(x):
    if x < 0:
        return int(x) - 1
    return int(x)

def _sign(x):
    if x < 0:
        return -1
    return 1


class ummy(Dfunc):
    max_dof = 10000 # any larger dof will be rounded to float('inf')
    nsig = 2 # The number of digits to quote the uncertainty to
    thousand_spaces = True # If true spaces will be placed between groups of three digits.
    
    # Set sci_notation to None to automatically switch to scientific notation
    # when the exponent is greater than sci_notation_high or less than 
    # sci_notation_low.  Or set sci_notation to True (or False) to force 
    # scientific notation to (not) be displayed if possible.
    sci_notation = None
    sci_notation_high = 7
    sci_notation_low = -3
    
    # if rounding_u is true then uncertianty will be added for floating point 
    #   rounding errors
    rounding_u = False
    
    max_digits = 15
    
    def __init__(self,x,u=0,dof=float('inf'),utype=None):
        if isinstance(x,ummy):
             self._copy(x,self,formatting=False)
             return
         
        try:
            float(x)
        except TypeError:
            raise TypeError('x must be a real number')
            
        try:
            float(u)
            if not isnan(u) and u < 0:
                raise TypeError()
        except TypeError:
            raise TypeError('u must be a real number >= 0')
                
        if self.rounding_u and not isinstance(x,Rational):
            finfo,x = _getfinfo(x)
            if finfo is _iinfo:
                warn('numpy.finfo cannot get the floating point accuracy for x\nno uncertainty will be included to account for floating point rounding errors',UncertiantyPrecisionWarning)
            else:
                uc = _combu(u,float(abs(x)*finfo.rel_u),0)
            
            uinfo = _getfinfo(u)[0]
            if uc < uinfo.warn_u and u is not 0:
                warn('a gummy/ummy was defined with an uncertainty too small to be accurately represented by a float\nuncertainty values may contain significant numerical errors',UncertiantyPrecisionWarning)
            u = uc
            self._finfo = finfo
        else:
            self._finfo = None
            
        if isinstance(x,Fraction):
            x = MFraction(x)
        if isinstance(u,Fraction):
            u = MFraction(u)
            
        self._x = x
        self._u = u
        
        self._refs = 1
        if not isinf(u) and not isnan(u) and u > 0:
            if isinstance(dof,_GummyRef):
                self._ref = dof
                self._refs = utype
            else:
                try:
                    if isinstance(dof,Integral):
                        dof = int(dof)
                    else:
                        dof = float(dof)
                    if dof <= 0 or isnan(dof):
                        raise TypeError()
                except TypeError:
                    raise TypeError('dof must be a real number > 0')
                if dof > self.max_dof:
                    dof = float('inf')
                    
                self._ref = _GummyRef(dof)
                if utype is not None:
                    GummyTag.set_tag(self,utype)
        else:
            self._ref = None
        
    @property
    def x(self):
        return self._x
        
    @property
    def u(self):
        return self._u
        
    @property
    def dof(self):
        """
        `float`, read-only

        Returns the number or degrees of freedom that the uncertainty of the 
        gummy is based on.  If the gummy was created as the result of an 
        operation between two or more other gummys, then the dof is the effective
        number of degrees of freedom calculated using a version of the Welch-
        Satterthwaite approximation that takes into account correlations, see
        [R. Willink, Metrologia, 44, 340 (2007)].
        """
        if self._ref is None:
            return float('inf')
        if self._ref.dof < 1:
            # Occasionally correlations can result in a _dof less than 1;
            # see the _get_dof method.
            return 1
        return self._ref.dof
    
    @property
    def _dof(self):
        if self._ref is None:
            return float('inf')
        return self._ref.dof
    @_dof.setter
    def _dof(self,v):
        if self._ref is not None:
            self._ref.dof = v
        
    @property
    def const(self):
        """`bool`, read-only
        
        Returns `True` if the gummy represents an exact value with no uncertainty.
        """
        return (self._u == 0)
    
    @property
    def finfo(self):
        if self._finfo is None:
            self._finfo = _getfinfo(self._x)
        return self._finfo
        
    @staticmethod
    def _copy(s,r,formatting=True,tofloat=False):
        # copies attributes of s to r
        if tofloat:
            r._x = float(s._x)
            r._u = float(s._u)
        else:
            r._x = s._x
            r._u = s._u
        r._ref = s._ref
        r._refs = s._refs
        r._finfo = s._finfo

        if formatting:
            if s.nsig != type(r).nsig:
                r.nsig = s.nsig
            if s.thousand_spaces != type(r).thousand_spaces:
                r.thousand_spaces = s.thousand_spaces
    
    def copy(self,formatting=True,tofloat=False):
        """
        Returns a copy of the gummy.  If the `formatting` parameter is
        `True` the display formatting information will be copied and if
        `False` the display formatting will be set to the default for a
        new gummy.  The default for `formatting` is `True`.  If tofloat
        is true the x and u properties will be converted to float values
        before copying.
        """
        r = type(self).__new__(type(self))
        self._copy(self,r,formatting=formatting,tofloat=tofloat)
        return r
        
    def tofloat(self):
        """
        Returns a copy of the gummy with x an u converted to floats.
        """
        return self.copy(formatting=False,tofloat=True)
    
    def correlation(self, g):
        """
        Returns the correlation coefficient between `self` and `g`."""
        if not isinstance(g,ummy):
            return 0
        if self._ref is None or g._ref is None:
            return 0
        return (self._refs*g._refs)*self._ref.cor(g._ref)
        
    def covariance(self,g):
        """
        Returns the covariance between `self` and `g`.
        """
        if not isinstance(g,ummy):
            return 0
        if self._ref is None or g._ref is None:
            return 0
        return self._u*g._u*(self._refs*g._refs)*self._ref.cor(g._ref)
        
    @staticmethod
    def correlation_matrix(gummys):
        """
        Returns the correlation matrix of a list or array of gummys.
        """
        return [[b.correlation(a) if isinstance(b,ummy) else 0 
                for b in gummys] for a in gummys]
        
    @staticmethod
    def covariance_matrix(gummys):
        """
        Returns the variance-covariance matrix of a list or array of gummys.
        """
        return [[b.covariance(a) if isinstance(b,ummy) else 0 for b in gummys] 
                for a in gummys]
                                
    @staticmethod
    def _set_correlation_matrix(gummys, matrix):
        n = len(gummys)
        m = np.asarray(matrix)
        
        if not np.allclose(m.T,m,atol = 10*_GummyRef._cortol):
            raise ValueError('the correlation matrix is not symmetric')
        e,v = np.linalg.eigh(m)
        if any(e <= -_GummyRef._cortol):
            raise ValueError('the correlation matrix is not positive semi-definate')
            
        if m.shape != (n,n):
            raise ValueError('matrix must have shape len(gummys) x len(gummys)')
        
        # Check that the diagonal elements are 1 (leaving room for rounding errors).
        for i in range(n):
            if abs(m[i][i] - 1.0) > 0.01:
                raise ValueError('correlation matrix diagonal elements must be 1')
            
        # Set the correlations using the off diagonal elements.
        for i in range(0,n):
            for j in range ((i+1),n):
                if m[i][j] != 0:
                    if gummys[i]._ref is not None and gummys[j]._ref is not None:
                        gummys[i]._ref.set_cor(gummys[j],m[i][j])
                    else:
                        raise ValueError('constants cannot have a non-zero correlation')
                    
    @staticmethod
    def _set_covariance_matrix(gummys, matrix):
        n = len(gummys)
        try:
            assert len(matrix) == n
            for r in matrix:
                assert len(r) == n
        except:
            raise ValueError('the covariance matrix must be a square matrix of the same length as gummys')
        
        # Set the u of each gummy using the diagonal of the matrix.
        for i in range(n):
            if matrix[i][i] == 0:
                gummys[i]._u = 0
                gummys[i]._ref = None
            else:
                if gummys[i]._ref is None:
                    gummys[i]._ref = _GummyRef()
                gummys[i]._u = matrix[i][i]**0.5
        
        # Set the correlations
        m = [[e/(gummys[i]._u*gummys[j]._u) for j,e in enumerate(r)]
             for i,r in enumerate(matrix)]
        ummy._set_correlation_matrix(gummys,m)
        
    @classmethod
    def create(cls,x,u=None,dof=None,correlation_matrix=None,
               covariance_matrix=None):
        """
        A class method that creates a list of (possibly) correlated ummys.
        
        Parameters
        ----------
        x:
            A list of floats corresponding to the x-value of each ummy.
        u, dof, k, loc, utype:  optional
            Lists that correspond to the
            parameters in the ummy initializer (with the i-th value in each
            list passed to the initializer for the i-th ummy).  These may also be a single
            value with this same value is to passed to each initializer.
        correlation_matrix:  optional
            A list or array to be used as the correlation
            matrix of the ummys. This is optional and must be set to the
            default value of None if the covariance_matrix is specified.
            If both the correlation_matrix and the covariance_matrix are
            None (or omitted) then the ummys will be uncorrelated.
        covariance_matrix:   optional
            A list or array to be used as the variance-
            covariance matrix of the ummys.  If the covariance matrix is
            specified the u parameter will be ignored  This parameter is
            optional and must be set to the default value of None if the
            correlation_matrix is specified.  If both the correlation_matrix
            and the covariance_matrix are None (or omitted) then the ummys
            will be uncorrelated.
                
        Returns
        -------
        a list of ummys
        """
       
            
        if covariance_matrix is not None:
            if correlation_matrix is not None:
                raise TypeError('correlation_matrix and covariance_matrix cannot both be specified')
                
        n = len(x)
        if u is None:
            u = [0]*n
        elif _isscalar(u):
            u = [u]*n
        if dof is None:
            dof = [float('inf')]*n
        elif _isscalar(dof):
            dof = [dof]*n
            
        ret = [cls(x[i],u=u[i],dof=dof[i]) for i in range(n)]
            
        if correlation_matrix is not None:
            cls._set_correlation_matrix(ret, correlation_matrix)
        
        if covariance_matrix is not None:
            cls._set_covariance_matrix(ret, covariance_matrix)
            
        return ret
        
    def __str__(self):
        x = float(self._x)
        try:
            xabs = abs(x)
           
            if xabs != 0:
                xexp = _floor(log10(xabs))
            else:
                xexp = None
                
            u = float(self.u)
            
            if self._u == 0 or isnan(self._u) or isinf(self._u):
                if xexp is None:
                    return '0'
                else:
                    if (((self.sci_notation is None and (xexp > self.sci_notation_high or xexp < self.sci_notation_low)) 
                        or self.sci_notation) ):
                        xtxt = self._format_mantissa('unicode',x*10**(-xexp),
                                                     self.finfo.precision)
                        etxt = 'e'
                        if xexp >= 0:
                            etxt += '+'
                        etxt += str(xexp)
                        return xtxt + etxt
                    else:
                        return self._format_mantissa('unicode',x,
                                                     self.finfo.precision)
                    
            uabs = abs(u)
            if uabs != 0:
                uexp = _floor(log10(uabs))
                
            if xexp is None:
                xp = uexp
            else:
                xp = max(xexp,uexp)
                
            psn = (uexp - self.nsig + 1 > 0)
            
            if (((self.sci_notation is None and (xp > self.sci_notation_high or xp < self.sci_notation_low)) 
                        or self.sci_notation) or psn or (xexp is not None and (uexp > xexp and xexp == 0))):
                xtxt = self._format_mantissa('unicode',x*10**(-xp),-uexp+xp+self.nsig-1)
                etxt = 'e'
                if xp >= 0:
                    etxt += '+'
                etxt += str(xp)
            else:
                xtxt = self._format_mantissa('unicode',x,-uexp+self.nsig-1)
                etxt = ''
                
            if self.nsig <= 0:
                return xtxt + etxt
    
            utxt = self._format_mantissa('unicode',u*10**(-uexp),self.nsig-1,parenth=True)
            return xtxt + '(' + utxt + ')' + etxt
        except:
            try:
                return(str(self.x) + '{' + str(self.u) + '}' + '??')
            except:
                try:
                    return(str(self.x) + '{??}')
                except:
                    return('??')
                
    def __repr__(self):
        return self.__str__()
          
    def _format_mantissa(self,fmt,x,sig,parenth=False):            
        try:
            if isnan(x):
                return 'nan'
            if isinf(x) and x > 0:
                if fmt == 'html':
                    return '&infin;'
                if fmt == 'latex':
                    return r'\infty'
                if fmt == 'ascii':
                    return 'inf'
                return '\u221E'
            if isinf(x) and x < 0:
                if fmt == 'html':
                    return '-&infin;'
                if fmt == 'latex':
                    return r'-\infty'
                if fmt == 'ascii':
                    return '-inf'
                return '-\u221E'
        except:
            pass
        
        ellipses = ''
        if isinstance(x,Rational) and sig is None:
            if x == int(x):
                s = str(int(x))
            else:
                xf = float(x)
                e = 15 - _floor(log10(abs(xf)))
                if e < 0:
                    e = 0
                n = 10**e*x
                if int(n) != n:
                    ellipses = '...'
                s = str(round(xf,e))
        elif sig is None:
            s = str(x)
        else:
            if x != 0 and self.max_digits is not None:
                if mpf is not None and isinstance(x,mpf):
                    em = _floor(mp.log10(abs(x)))
                else:    
                    em = _floor(log10(abs(float(x))))
                e = self.max_digits - em
                if e < 0:
                    e = 0
                if sig > e and sig > 0:
                    ellipses = '...'
                    sig = e
            if sig < 0:
                sig = 0
            if mpf is not None and isinstance(x,mpf):
                s = mp.nstr(x,n=sig+1,strip_zeros=False)
            else:
                s = round(float(x),sig)
                s = ('{:.' + str(sig) + 'f}').format(s)
        
        if parenth:
            s = s.replace('.','')
            s = s.lstrip('0')
            if len(s) == 0:
                s = '0'
            return s
        elif self.thousand_spaces:
            # Now insert spaces every three digits
            if fmt == 'html':
                sp = '&thinsp;'
            elif fmt == 'latex':
                sp = r'\,'
            else:
                sp = ' '
                
            d = s.find('.')
            if d != -1:
                i = d + 1
                while i < len(s) and s[i].isdigit():
                    i += 1
                nr = i - d - 1
            else:
                nr = 0
                d = len(s)
                
            i = d - 1
            while i > 0 and s[i].isdigit():
                i -= 1
            nl = d - i - 1
            
            if nr <= 4 and nl <= 4:
                return s + ellipses
                
            s = list(s)    
            if nr > 4:
                for i in range(int(nr/3)):
                    s.insert(d + i + (i+1)*3 + 1, sp)
            if nl > 4:
                for i in range(int(nl/3)):
                    s.insert(d - (i+1)*3, sp)
                    
            s = ''.join(s)
            if s.endswith(sp):
                s = s[:-(len(sp))]
                
        return s + ellipses
        
    @classmethod
    def _get_dof(cls,dof1,dof2,du1,du2,c):
        # Use the Welch-Satterhwaite approximation to combine dof1 and dof2.
        # See [R. Willink, Metrologia, 44, 340 (2007)] concerning the use
        # of the Welch-Satterwaite approximation with correlated values.
        # Note that we allow the return value to be less than one,
        # but when the dof is retrieved using the dof property get method, 
        # values less than one will be rounded up to one.
            
        # Willink's formula, [R. Willink, Metrologia, 44, 340 (2007)], for
        # the W-S approximations for correlated uncertainties is:
        # d = (du1**2 + c*du1*du2)**2/dof1 + (du2**2 + c*du1*du2)**2/dof2
        # Willink's formula is used in the _apply method but not here.
        
        d = du1**4/dof1 + du2**4/dof2
        d += 2*c**2*(((du1 + du2)**4 - du1**4 - du2**4)/
                     ((1 - 2*c + 2*c**2)*(dof1+dof2)))

        if d == 0:
            return float('inf')
        
        r = 1/d
        if r > cls.max_dof:
            return float('inf')
        
        if r < -1e-6:
            raise ValueError('dof is negative')
        if r < 0:
            r == 1
        
        return r
    
    @classmethod
    def _apply(cls,function,derivative,*args,fxdx=None):
        if fxdx is None:
            x = [a.x if isinstance(a,ummy) else a for a in args]
            fx = function(*x)
            d = derivative(*x)
        else:
            fx,d,x = fxdx
        if len(args) == 1:
            d = [d]
            
        ad = [(a,a.u*p) for a,p in zip(args,d) 
              if isinstance(a,ummy) and a._u != 0]
        if len(ad) == 0:
            return cls(fx)
        args,du = list(map(list, zip(*ad)))
            
        if len(args) == 1:
            refs = -args[0]._refs if du[0] < 0 else args[0]._refs
            r = cls(fx,u=abs(du[0]),dof=args[0]._ref,utype=refs)
            return r
        
        maxdu = max(abs(a) for a in du)
        if maxdu == 0:
            return cls(fx)
        dun = [d/maxdu for d in du]
        
        u = sum(d**2 for d in dun)
        u += 2*sum(args[i].correlation(args[j])*dun[i]*dun[j] 
                 for i in range(len(args)) for j in range(i+1,len(args)))
        
        if not isnan(u):
            if u < 0:
                if u < -1e6:
                    # something horrible happened, should only get here due to a bug
                    raise FloatingPointError('u is sqrt of ' + str(maxdu**2*u))
                u = 0
                
            u = u**0.5
            u = maxdu*u
            
        if u == 0 or isnan(u):
            return cls(fx)
        
        du = [d/u for d in du]
        dof = float('inf')
        if any(not isinf(a._dof) for a in args):
            dm = sum(sum(args[i].correlation(args[j])*du[i]*du[j] 
                         for i in range(len(du)))**2/args[j]._dof 
                         for j in range(len(du)))
            if dm > 0:
                dof = 1/dm
                
        r = cls(fx,u=u,dof=dof)
        
        rl = {a._ref for a in args}
        for i,a in enumerate(args):
            s = sum(a.correlation(args[j])*du[j] for j in range(len(du)))
            a._ref.combl(r,float(a._refs*s),float(a._refs*du[i]),rl)
            if r._ref in rl:
                break
        r._check_cor()
            
        return r
        
    @classmethod
    def _napply(cls,function,*args,fxx=None):
        n = len(args)
        if n == 0:
            return cls(function(),0)
        
        d = _der(function,*args)
        
        if fxx is not None:
            fxx = (fxx[0],d,fxx[1])
        
        if n == 1:
            return cls._apply(function,lambda x: d,args[0],fxdx=fxx)

        return cls._apply(function,lambda *x: d,*args,fxdx=fxx)
    
    def _check_cor(self):
        if self._ref is None:
            return
        
        rl = list(self._ref._cor.items())
        for k,v in rl:
            if k is not None:
                if abs(v) > 1.01:
                    raise ValueError('abs(correlation) > 1')
                if abs(v) < _GummyRef._cortol:
                    del k._cor[self._ref]
                    del self._ref._cor[k]
                    return
                if v > _GummyRef._cortolp:
                    k.copyto(self,1)
                    return
                if v < _GummyRef._cortoln:
                    k.copyto(self,-1)
                    return
                
    def _combc(self,a,b,dua,dub,c):
        dua = float(dua)
        dub = float(dub)

        a._ref.combl(self,c*dub*a._refs + dua*a._refs,dua*a._refs)
        if self._ref is not a._ref:
            b._ref.combl(self,c*dua*b._refs + dub*b._refs,dub*b._refs,[a._ref])
            if self._ref is not b._ref:
                self._check_cor()   
                
    def _add(self,b):
        if not isinstance(b,ummy):
            return type(self)(self._x + b,self._u,dof=self._ref,utype=self._refs)
    
        c = self.correlation(b)
        x = self._x + b._x
        u = _combu(self._u,b._u,c)
        
        if u == 0:
            return type(b)(x)
        
        if self._u == 0:
            return type(b)(x,u,dof=b._ref,utype=b._refs)
        
        if b._u == 0:
            return type(b)(x,u,dof=self._ref,utype=self._refs)
        
        dua = self._u/u
        dub = b._u/u
        if not isinf(self._ref.dof) or not isinf(b._ref.dof):
            dof = self._get_dof(self._dof,b._dof,dua,dub,c)
        else:
            dof = float('inf')
            
        r = type(b)(x,u=u,dof=dof)
    
        r._combc(self,b,dua,dub,c)
        
        return r
    
    def _radd(self,b):
        return type(self)(self._x + b,self._u,dof=self._ref,utype=self._refs)
    
    def _sub(self,b):           
        if not isinstance(b,ummy):
            return type(self)(self._x - b,self._u,dof=self._ref,utype=self._refs)
            
        c = self.correlation(b)
        x = self._x - b._x
        u = _combu(self._u,-b._u,c)
            
        if u == 0:
            return type(b)(x)
        
        if self._u == 0:
            return type(b)(x,u,dof=b._ref,utype=-b._refs)
        
        if b._u == 0:
            return type(b)(x,u,dof=self._ref,utype=self._refs)
        
        dua = self._u/u
        dub = -b._u/u
        if not isinf(self._ref.dof) or not isinf(b._ref.dof):
            dof = self._get_dof(self._dof,b._dof,dua,dub,c)
        else:
            dof = float('inf')
        
        r = type(b)(x,u=u,dof=dof)
        
        r._combc(self,b,dua,dub,c)

        return r
    
    def _rsub(self,b):
        r = type(self)(b - self._x,self._u,dof=self._ref,utype=-self._refs)
        return r
    
    def _mul(self,b):
        if not isinstance(b,ummy):
            refs = -self._refs if b < 0 else self._refs
            return type(self)(self._x*b,abs(self._u*b),dof=self._ref,utype=refs)
                                        
        c = self.correlation(b)
        x = self._x*b._x
        dua = b._x*self._u
        dub = self._x*b._u
        u = _combu(dua,dub,c)
            
        if u == 0:
            return type(b)(x)
        
        if self._u == 0:
            refs = -b._refs if self._x < 0 else b._refs
            return type(b)(x,u,dof=b._ref,utype=refs)
        
        if b._u == 0:
            refs = -self._refs if b._x < 0 else self._refs
            return type(b)(x,u,dof=self._ref,utype=refs)
        
        dua = dua/u
        dub = dub/u
        if not isinf(self._ref.dof) or not isinf(b._ref.dof):
            dof = self._get_dof(self._dof,b._dof,dua,dub,c)
        else:
            dof = float('inf')
            
        r = type(b)(x,u=u,dof=dof)
        
        r._combc(self,b,dua,dub,c)

        return r
    
    def _rmul(self,b):
        refs = -self._refs if b < 0 else self._refs
        return type(self)(self._x*b,abs(self._u*b),dof=self._ref,utype=refs)
    
    def _truediv(self,b):   
        if not isinstance(b,ummy):
            if b == 0:
                raise ZeroDivisionError('division by zero')
            if (isinstance(self._x,Integral) and 
                isinstance(b,Integral)):
                x = MFraction(self._x,b)
            else:
                x = self._x/b
            refs = -self._refs if b < 0 else self._refs
            return type(self)(x,abs(self._u/b),dof=self._ref,utype=refs)
                
        c = self.correlation(b)
        if b._x == 0:
            raise ZeroDivisionError('division by zero')
        else:
            if (isinstance(self._x,Integral) and 
                isinstance(b._x,Integral)):
                x = MFraction(self._x,b._x)
            else:
                x = self._x/b._x
            
        dua = self._u/b._x
        dub = -(self._x*b._u/b._x)/b._x
        
        u = _combu(dua,dub,c)
        
        if u == 0:
            return type(b)(x)
        
        if self._u == 0:
            refs = b._refs if self._x < 0 else -b._refs
            return type(b)(x,u,dof=b._ref,utype=refs)
        
        if b._u == 0:
            refs = -self._refs if b._x < 0 else self._refs
            return type(b)(x,u,dof=self._ref,utype=refs)
        
        dua = dua/u
        dub = dub/u
        if not isinf(self._ref.dof) or not isinf(b._ref.dof):
            dof = self._get_dof(self._dof,b._dof,dua,dub,c)
        else:
            dof = float('inf')
            
        r = type(b)(x,u=u,dof=dof)
        
        r._combc(self,b,dua,dub,c)

        return r
    
    def _rtruediv(self,b):
        if self._x == 0:
            raise ZeroDivisionError('division by zero')
        else:
            if (isinstance(self._x,Integral) and 
                isinstance(b,Integral)):
                x = MFraction(b,self._x)
            else:
                x = b/self._x
        u = abs(b*self._u/self._x**2)
        refs = self._refs if b < 0 else -self._refs
        return type(self)(x,u,dof=self._ref,utype=refs)
    
    def _pow(self,b):
        if isinstance(b,ummy) and b._u == 0:
            b = b._x
        if not isinstance(b,ummy):
            if b == 0:
                    return type(self)(1)
            if (isinstance(self._x,Integral) and 
                isinstance(b,Integral) and b < 0):
                x = MFraction(1,self._x**-b)
            else:
                x = self._x**b
            u = abs(b*self._x**(b-1)*self._u)
            if self._x < 0:
                sgn = int(-((-1)**b))
            else:
                sgn = 1
            if b < 0:
                sgn = -sgn
            return type(self)(x,u,dof=self._ref,utype=sgn*self._refs)

        if self._x <= 0:
            raise ValueError('a negative or zero value cannot raised to a power which has an uncertainty')            
                
        c = self.correlation(b)
        if (isinstance(self._x,Integral) and 
                isinstance(b._x,Integral) and b._x < 0):
            x = MFraction(1,self._x**-b._x)
        else:
            x = self._x**b._x
        dua = b._x*self._x**(b._x-1)*self._u
        lgx = log(self._x)
        
        # don't exactly remember why I did this,
        # but it causes problems if x is int.
        #try:
            #lgx = type(self._x)(lgx)
        #except:
            #pass
            
        dub = lgx*self._x**b._x*b._u
        u = _combu(dua,dub,c)
        
        if u == 0:
            return type(b)(x)
        
        if self._u == 0:
            refs = -b._refs if self._x < 0 else b._refs
            return type(b)(x,u,dof=b._ref,utype=refs)
        
        dua = dua/u
        dub = dub/u
        
        if not isinf(self._ref.dof) or not isinf(b._ref.dof):
            dof = self._get_dof(self._dof,b._dof,dua,dub,c)
        else:
            dof = float('inf')
            
        r = type(b)(x,u=u,dof=dof)
        
        r._combc(self,b,dua,dub,c)

        return r
    
    def _rpow(self,b):
        if b == 0:
            return type(self)(0)
        if (isinstance(self._x,Integral) and 
            isinstance(b,Integral) and self._x < 0):
            x = MFraction(1,b**-self._x)
        else:
            x = b**self._x
            
        if self._u == 0:
            type(self)(x)
            
        if b < 0:
            raise ValueError('a negative number cannot raised to a power which has an uncertainty')
            
        lgb = log(b)
        
        # don't exactly remember why I did this,
        # but it causes problems if x is int.
        #try:
            #lgb = type(b)(lgb)
        #except:
            #pass
            
        u = abs(b**self._x*lgb*self._u)
        refs = -self._refs if b < 0 else self._refs
        return type(self)(x,u,dof=self._ref,utype=refs)
    
    def _nprnd(self,f):
        return type(self)(f(self._x))
        
    def _floordiv(self,b):
        return self._truediv(b)._nprnd(_floor)
        
    def _rfloordiv(self,b):
        return self._rtruediv(b)._nprnd(_floor)
        
    def _mod(self,b):
        ret = ummy._apply(lambda x1,x2: x1%x2,
                          lambda x1,x2: (1, _sign(x2)*abs(x1//x2)),self,b)
        return type(self)(ret)
        
    def _rmod(self,b):
        ret = ummy._apply(lambda x1,x2: x1%x2,
                          lambda x1,x2: (1, _sign(x2)*abs(x1//x2)),b,self)
        return type(self)(ret)
        
    def __neg__(self):
        r = self.copy(formatting=False)
        r._x = -self._x
        if self._ref is not None:
            r._refs = -self._refs
        return r
        
    def __pos__(self):
        return self.copy(formatting=False)
        
    def __abs__(self):
        r = self.copy(formatting=False)
        if self._x < 0:
            r._x = -self._x
            if self._ref is not None:
                r._refs = -self._refs
        return r
    
    def __eq__(self,v):
        if self is v:
            return True
        if isinstance(v,ummy):
            if self._u == 0 and v._u == 0:
                return self._x == v._x
            if self._ref is v._ref and self._refs == v._refs:
                if self._x == v._x and self._u == v._u:
                    return True
            return False
        elif self._u == 0:
            return self._x == v
        else:
            return False
        
    def __float__(self):
        return float(self.x)
        
    def __int__(self):
        return int(self.x)
        
    def __complex__(self):
        return complex(self.x)
    
    @property
    def real(self):
        return self
    
    @property
    def imag(self):
        return type(self)(0)
    
    def conjugate(self):
        return self.copy(formatting=False)
    
    def angle(self):
        if self._x >= 0:
            return type(self)(0)
        else:
            return type(self)(pi)
        
    @property
    def utype(self):
        """
        `str` or `None`

        An arbitrary string value labeling the uncertainty type.
        """
        if self._ref is None or self._ref._tag is None:
            return None
        return self._ref._tag.name
        
    @staticmethod
    def _toummylist(x):
        if isinstance(x,ummy):
           return [x]
        elif isinstance(x,GummyTag):
            return x.get_values()
        elif isinstance(x,str):
            t = GummyTag.tags.get(x)
            if t is None:
                return []
            x = t.get_values()
            return x
        
        xl = []
        for e in x:
            if isinstance(e,ummy):
               xl.append(e)
            if isinstance(e,GummyTag):
                xl += e.get_values()
            if isinstance(e,str):
                t = GummyTag.tags.get(e)
                if t is not None:
                    xl += t.get_values()
        return xl
        
        
    def ufrom(self,x):
        """
        Gets the standard uncertainty contributed from particular gummys
        or utypes if all other free variables are held fixed.
        
        Parameters
        ----------
        x:  `gummy`, `str`, or array_like
            A gummy, a string referencing a utype or a list containing
            gummys and strings.
            
        Returns
        -------
        `float`
        
        Example
        -------
        >>>  a = gummy(1.2,0.2,utype='A')
        >>>  b = gummy(3.2,0.5,utype='A')
        >>>  c = gummy(0.9,0.2,utype='B')
        >>>  d = a + b + c
        >>>  d.ufrom('A')
        0.53851648071345048
        """
        x = ummy._toummylist(x)
        x = [i for i in x if self.correlation(i) != 0]
        if len(x) == 0:
            return 0

        v = ummy.correlation_matrix(x)
            
        b = [self.correlation(z) for z in x]
        s = np.linalg.lstsq(v,b,rcond=None)[0]
        u = 0
        
        d = [i*self._u/j._u for i,j in zip(s,x)]
        for i in range(len(x)):
            for j in range(len(x)):
                u += d[i]*d[j]*x[i].correlation(x[j])*x[i]._u*x[j]._u
                
        return u**0.5
        
    def doffrom(self,x):
        """
        Gets the degrees of freedom contributed from particular gummys or
        utypes if all other free variables are held fixed.
        
        Parameters
        ----------
        x:  `gummy`, `str`, or array_like
            A gummy, a string referencing a utype or a list containing
            gummys and strings.

        Returns
        -------
        `float`

        Example
        -------
        >>>  a = gummy(1.2,0.2,dof=5,utype='A')
        >>>  b = gummy(3.2,0.5,dof=7,utype='A')
        >>>  c = gummy(0.9,0.2,utype='B')
        >>>  d = a + b + c
        >>>  d.doffrom('A')
        9.0932962619709627
        """
        x = ummy._toummylist(x)
        x = [i for i in x if self.correlation(i) != 0]
        if len(x) == 0:
            return float('inf')
            
        v = ummy.correlation_matrix(x)
        b = [self.correlation(z) for z in x]
        s = np.linalg.lstsq(v,b,rcond=None)[0]
        d = [i*self._u/j._u for i,j in zip(s,x)]
        usq = 0
        dm = 0
        for i in range(len(x)):
            for j in range(len(x)):
                usqi = d[i]*d[j]*x[i].correlation(x[j])*x[i]._u*x[j]._u
                usq += usqi
                dm += usqi**2/x[i].dof
                
        if dm == 0:
            return float('inf')
        dof = usq**2/dm
        if dof > ummy.max_dof:
            return float('inf')
        if dof < 1:
            return 1
        return dof
        
        
class _GummyRef:
    # Instances of this class are hashable unlike gummys, so we can use this in 
    # a gummy's ._cor weak key dictionary and a GummyTag's .values weak set.

    _cortol = 1e-8 # correlations smaller than this are rounded to zero
    _cortolp = 1 - 1e-8 # correlations larger than this are rounded to 1
    _cortoln = -1 + 1e-8 # correlations smaller than this are rounded to -1
    
    def __init__(self,dof=float('inf')):
        self._cor = weakref.WeakKeyDictionary()
        self._tag = None
        self._tag_refs = []
        self._tag_deps = []
        self.dof = dof
        
    def cor(self,ref):
        if ref is self:
            return 1
        c = self._cor.get(ref)
        if c is None:
            return 0
        else:
            return c
            
    def set_cor(self,g,c):
        # We allow input values slighly greater than 1 without raising an 
        # exception to give room for rounding errors (but values greater than
        # 1 will be rounded to 1 below).
        
        if abs(c) > 1.01:
            raise ValueError('abs(c) > 1')
            
        if abs(c) < _GummyRef._cortol:
            if g._ref in self._cor:
                del self._cor[g._ref]
                del g._ref._cor[self]
            return
        
        if c > _GummyRef._cortolp:
            self.copyto(g,1)
            return
            
        if c < _GummyRef._cortoln:
            self.copyto(g,-1)
            return
        
        # We can't let an ummy with a utype die and be collected with the garbage
        # if it is correlated with a living ummy in case ufrom() or doffrom() is 
        # called with the utype as an argument.  So we add a strong reference 
        # pointing to the utyped ummy to the GummyRef of all other ummys that 
        # are correlated with it.
        if self._tag is not None:
            g._ref._tag_deps.append(self.get_tag_ref())
        elif len(self._tag_deps) > 0:
            g._ref._tag_deps += self._tag_deps
        
        self._cor[g._ref] = c
        g._ref._cor[self] = c
            
    def add_cor(self,g,c):
        if c == 0:
            return
        v = self._cor.get(g._ref)
        if v is None:
            v = c
        else:
            v += c
        self._cor[g._ref] = v
        g._ref._cor[self] = v
        
    @staticmethod
    def check_cor(r):
        a = list(r._ref._cor.items())
        for k,v in a:
            if k is not None:
                if abs(v) > 1.01:
                    raise ValueError('abs(correlation) > 1')
                if abs(v) < _GummyRef._cortol:
                    del k._cor[r._ref]
                    del r._ref._cor[k]
                    return
                if v > _GummyRef._cortolp:
                    k.copyto(r,1)
                    return
                if v < _GummyRef._cortoln:
                    k.copyto(r,-1)
                    return
                
    def copyto(self,g,refs=1):
        g._ref = self
        g._refs = refs
        if self._tag is not None:
            self._tag_refs.append(weakref.ref(g))
            
    def combl(self,g,c1,c2,rl=None):
        self.set_cor(g,c1)
        if g._ref is not self:
            a = list(self._cor.items())
            for k,v in a:
                if k is not None and k is not g._ref:
                    if rl is None or not any(i is k for i in rl):
                        k.add_cor(g,c2*v)
                        
    def get_tag_ref(self):
        try:
            ret = self._tag_refs[-1]()
            while ret is None:
                # The GummyRef may be shared by several copys of the original
                # ummy.  Find one that is still alive.
                self._tag_refs.pop()
                ret = self._tag_refs[-1]()
            return ret
        except IndexError:
            return None
            

class GummyTag:
    # Do not create GummyTag objects directly.  A GummyTag instance is created
    # each time a gummy instance with a new and unique utype parameter is created.

    tags = weakref.WeakValueDictionary()
        
    @staticmethod
    def set_tag(g,tag):
        if isinstance(tag,str):
            tag = tag.strip()
            if tag in GummyTag.tags:
                tag = GummyTag.tags[tag]
            else:
                tag = GummyTag(tag)

        elif not isinstance(tag,GummyTag):
            raise ValueError('utype must be either None or a str')
            
        tag.values.add(g._ref)
        g._ref._tag = tag
        g._ref._tag_refs.append(weakref.ref(g))
        
    def __init__(self,tag_name):
        GummyTag.tags[tag_name] = self
        
        self.values = weakref.WeakSet()
        
        self.name = tag_name
        
    def get_values(self):
        vals = list(self.values)

        # Can't put this in a list comprehension since v.ref is a weakref and
        # it could die in the middle of the loop.
        ret = []
        for v in vals:
            if v is not None:
                r = v.get_tag_ref()
                if r is not None:
                    ret.append(r)
                
        return ret
    
                        
def _der(function,*args):
    # computes the numerical derivative of function with respect to args.
    # puts zeros for any of args that are not an ummy
    n = len(args)
    if n == 1:
        arg = args[0]
        if not isinstance(arg,ummy) or arg._u == 0:
            return 0
                
        df = None
        a = np.empty([8,8])
        x = arg.x
        s = 2*arg.u
        a[0,0] = (function(x+s)-function(x-s))/(2*s)
        for k in range(1,8):
            s = s/1.4
            a[0,k] = (function(x+s)-function(x-s))/(2*s)
            w = 1.96
            for j in range(1,k):
                a[j,k] = (a[j-1,k]*w-a[j-1,k-1])/(w-1)
                w = 1.96*w
                dff = max(abs(a[j,k]-a[j-1,k]),abs(a[j,k]-a[j-1,k-1]))
                if df is None or dff <= df:
                    df = dff
                    d = a[j,k]
            if df is not None and abs(a[k,k]-a[k-1,k-1]) > 2*df:
                break
            
        return d
        
    v = np.empty(n)
    for i,a in enumerate(args):
        if isinstance(a,ummy):
            v[i] = a.x
        else:
            v[i] = a
    d = np.zeros(n)
    for i,p in enumerate(args):
        if isinstance(p,ummy) and p._u > 0:          
            df = None
            s = 2*p.u
            x1 = np.array(v)
            x1i = x1[i]
            x1[i] = x1i+s
            x2 = np.array(v)
            x2i = x2[i]
            x2[i] = x2i-s
            a = np.empty([8,8])
            a[0,0] = (function(*x1)-function(*x2))/(2*s)
            for k in range(1,8):
                s = s/1.4
                x1[i] = x1i+s
                x2[i] = x2i-s
                a[0,k] = (function(*x1)-function(*x2))/(2*s)
                w = 1.96
                for j in range(1,k):
                    a[j,k] = (a[j-1,k]*w-a[j-1,k-1])/(w-1)
                    w = 1.96*w
                    dff = max(abs(a[j,k]-a[j-1,k]),abs(a[j,k]-a[j-1,k-1]))
                    if df is None or dff <= df:
                        df = dff
                        di = a[j,k]
                if df is not None and abs(a[k,k]-a[k-1,k-1]) > 2*df:
                    break
            d[i] = di
            
    return d
        
class MFraction(Fraction):
    """
    A fraction.Fraction sub-class that works with mpmath.mpf objects
    """
    def _mpmath_(self,p,r):
        return rational.mpq(self.numerator,self.denominator)
